"""
fastgit（简称 fit） 是 fitlog 的一部分，用于帮助用户进行自动 commit。自动 commit 的文件在 .fitlog 文件夹中，与 .git 文件夹类似。
用户需要使用 fitlog 命令行工具初始化项目，项目默认会生成一个名为 .fitconfig 的配置文件，可以配置监控文件的规则等参数。

"""
from .committer import Committer

committer = Committer()



import logging
import os
from datetime import datetime
import argparse
import json
import time
import re
from typing import Union, Dict
from configparser import ConfigParser

from .log_read import is_dirname_log_record
from ..fastgit import committer
import warnings

def _check_debug(func):
    """
    函数闭包，只有非 debug 模式才会执行原始函数
    
    :param func: 原始函数，函数的第一个参数必须为 Logger 对象
    :return: 加上闭包后的函数
    """
    
    def wrapper(*args, **kwargs):
        if args[0]._debug:
            return
        else:
            return func(*args, **kwargs)
    
    return wrapper


def _check_log_dir(func):
    """
    函数闭包，检查原始函数执行所需的条件是否满足，只有满足才会执行
    
    1 如果没有initialize, 说明还没有设置
    
    2 如果default_log_dir不为None，设置使用default_log_dir调用set_log_dir
    
    3 否则报错
    
    :param func: 原始函数，函数的第一个参数必须为 Logger 对象
    :return: 加上闭包后的函数
    """
    
    def wrapper(*args, **kwargs):
        if not args[0].initialized and args[0].default_log_dir is not None:
            args[0].set_log_dir(args[0].default_log_dir)
        elif not args[0].initialized:
            raise RuntimeError("You have to call `logger.set_log_dir()` to set where to _save log first.")
        return func(*args, **kwargs)
    
    return wrapper


class Logger:
    """
    用于处理日志的类，fitlog 的核心
    """
    
    def __init__(self):
        self.initialized = False
        self.save_on_first_metric_or_loss = True
        self.default_log_dir = None
        self._cache = []
        self._debug = False
        self.fit_id = None
        self.git_id = None
        
        self._log_dir = None
    
    def debug(self):
        """
        再引入logger之后就调用，本次运行不会记录任何东西。所有函数无任何效用
        
        :return:
        """
        self._debug = True
    
    @_check_debug
    def commit(self, file: str, fit_msg: str = None):
        """
        调用 committer.commit 进行自动 commit

        :param file: 以该路径往上寻找.fitlog所在文件夹。一般传入__file__即可
        :param fit_msg: 针对该实验的说明
        :return:
        """
        msg = committer.commit(file=file, commit_message=fit_msg)
        if msg['status'] == 0:  # 成功了
            config = committer.config
            self.fit_id = msg['msg']
            self.save_on_first_metric_or_loss = config.getboolean('log_settings', 'save_on_first_metric_or_loss')
            self.default_log_dir = os.path.join(committer.work_dir, config.get('log_settings', 'default_log_dir'))
        else:
            raise RuntimeError("It seems like you are not running under a folder governed by fitlog.\n" + msg['msg'])
    
    @_check_debug
    def set_save_on_first_metric_or_loss(self, flag: bool = True):
        """
        是否只在 metric操作后创建 log 文件
        
            1 如果为False, 只要有任何add_*方法都会创建log文件，但大多数时候这都是不必要的。
            
            2 如果为True, 只有在第一次存入metric或者loss时才会真正创建log。
        
        :param flag:
        :return:
        """
        assert isinstance(flag, bool)
        self.save_on_first_metric_or_loss = flag
    
    @_check_debug
    def set_log_dir(self, log_dir: str, new_log: bool=False):
        """
        设定log 文件夹的路径，在进行其它操作前必须先指定日志路径

        :param log_dir: log 文件夹的路径
        :param new_log: 是否开始新的一条log记录
        :return:
        """
        if new_log:
            self._clear()

        if self.initialized:
            if self._log_dir != log_dir:
                raise RuntimeError("Don't set log dir again.")
            else:
                return
        
        if not os.path.exists(log_dir):
            raise NotADirectoryError("`{}` is not exist.".format(log_dir))
        if not os.path.isdir(log_dir):
            raise FileExistsError("`{}` is not a directory.".format(log_dir))
        
        # prepare file directory
        if is_dirname_log_record(log_dir):
            warnings.warn("Append to an already exist log.")
            self._log_dir = os.path.dirname(log_dir)
            self._save_log_dir = log_dir
        else:
            self._log_dir = log_dir
        
        if not os.access(log_dir, os.W_OK):
            raise PermissionError("write is not allowed in `{}`. Check your permission.".format(log_dir))
        
        self.initialized = True

    def _clear(self):
        """
        内部函数，将logger置为未初始化
        :return:
        """
        self._save()
        self.initialized = False
        self._cache = []
        for attr_name in ['_save_log_dir', '_log_dir', 'total_steps']:
            if hasattr(self, attr_name):
                delattr(self, attr_name)

        for logger_name in ['meta_logger', 'hyper_logger', 'metric_logger', 'other_logger', 'progress_logger',
                          'loss_logger']:
            if hasattr(self, logger_name):
                logger = getattr(self, logger_name)
                handlers = logger.handlers[:]
                for handler in handlers:
                    handler.close()
                    handler.flush()
                    logger.removeHandler(handler)
                delattr(self, logger_name)

    def _create_log_files(self):
        """
        创建日志文件
        """
        if not hasattr(self, 'meta_logger'):
            if not hasattr(self, '_save_log_dir'):
                now = datetime.now().strftime('%Y%m%d_%H%M%S')
                self._save_log_dir = os.path.join(self._log_dir, 'log_' + now)
                while os.path.exists(self._save_log_dir):
                    time.sleep(1)
                    now = datetime.now().strftime('%Y%m%d_%H%M%S')
                    self._save_log_dir = os.path.join(self._log_dir, 'log_' + now)
                os.mkdir(self._save_log_dir)
            # prepare logger
            self.meta_logger = logging.getLogger('meta')
            self.hyper_logger = logging.getLogger('hyper')
            self.metric_logger = logging.getLogger('metric')
            self.other_logger = logging.getLogger('other')
            self.loss_logger = logging.getLogger('loss')
            self.progress_logger = logging.getLogger('progress')
            
            formatter = logging.Formatter('%(message)s')  # 只保存记录的时间与记录的内容
            meta_handler = logging.FileHandler(os.path.join(self._save_log_dir, 'meta.log'), encoding='utf-8')
            hyper_handler = logging.FileHandler(os.path.join(self._save_log_dir, 'hyper.log'), encoding='utf-8')
            metric_handler = logging.FileHandler(os.path.join(self._save_log_dir, 'metric.log'), encoding='utf-8')
            loss_handler = logging.FileHandler(os.path.join(self._save_log_dir, 'loss.log'), encoding='utf-8')
            other_handler = logging.FileHandler(os.path.join(self._save_log_dir, 'other.log'), encoding='utf-8')
            progress_handler = logging.FileHandler(os.path.join(self._save_log_dir, 'progress.log'), encoding='utf-8')
            
            for handler in [meta_handler, hyper_handler, metric_handler, other_handler, loss_handler, progress_handler]:
                handler.setFormatter(formatter)
            
            for _logger, _handler in zip([self.meta_logger, self.hyper_logger, self.metric_logger, self.other_logger,
                                          self.loss_logger, self.progress_logger],
                                         [meta_handler, hyper_handler, metric_handler, other_handler, loss_handler,
                                          progress_handler]):
                _handler.setLevel(logging.INFO)
                _logger.setLevel(logging.INFO)
                _logger.addHandler(_handler)
            
            self.__add_meta()
    
    @_check_debug
    @_check_log_dir
    def __add_meta(self):
        """
        logger自动调用此方法添加meta信息
        """
        fit_id = None
        fit_msg = None
        if committer.last_commit is not None:
            fit_id = committer.last_commit[0]
            fit_msg = committer.last_commit[1]
        git_id = None
        git_msg = None
        res = committer.git_last_commit(self._log_dir)
        if res['status'] == 0:
            git_id = res['msg'][0]
            git_msg = res['msg'][1]
            self.git_id = res['msg'][0]
        
        _dict = {}
        for value, name in zip([fit_id, git_id, fit_msg, git_msg],
                               ['fit_id', 'git_id', 'fit_msg', 'git_msg']):
            if value is not None:
                if 'id' in name:
                    value = value[:8]
                _dict[name] = value
        
        _dict["state"] = 'running'
        _dict = {'meta': _dict}
        self._write_to_logger(json.dumps(_dict), 'meta_logger')
    
    @_check_debug
    @_check_log_dir
    def finish(self, status:int=0):
        """
        使用此方法告知 fitlog 你的实验已经正确结束。你可以使用此方法来筛选出失败的实验。

        :param int status: 告知当前实验的状态。0: 结束了; 1: 发生了错误
        """
        if status not in (0, 1):
            raise ValueError("status only supports 0,1 to stand for 'finish','error'.")
        if hasattr(self, 'meta_logger'):
            if status==0:
                _dict = {'meta': {'state': 'finish'}}
            else:
                _dict = {'meta': {'state': 'error'}}
            self._write_to_logger(json.dumps(_dict), 'meta_logger')
    
    @_check_debug
    @_check_log_dir
    def add_best_metric(self, value: Union[int, str, float, dict], name: str = None):
        """
        用于添加最好的 metric 。用此方法添加的值，会被显示在 metric 这一列中。

        :param value: 类型为 int, float, str, dict中的一种。如果类型为 dict，它的键的类型只能为 str，
                它的键值的类型可以为int, float, str 或符合同样条件的 dict
        :param name: 如果你传入的 value 不是字典，你需要传入 value 对应的名字。
        
        .. warning ::
            如果你在同时记录多个数据集上的performance, 请注意使用不同的名称进行区分

        """
        _dict = _parse_value(value, name=name, parent_name='metric')
        
        self._write_to_logger(json.dumps(_dict), 'metric_logger')
    
    @_check_debug
    @_check_log_dir
    def add_metric(self, value: Union[int, str, float, dict], step: int, name: str = None, epoch: int = None):
        """
        用于添加 metric 。用此方法添加的值，会被记录在 metric 这一列中
        
        :param value: 类型为 int, float, str, dict中的一种。如果类型为 dict，它的键的类型只能为 str，
                它的键值的类型可以为int, float, str 或符合同样条件的 dict
        :param step: 用于和 loss 对应的 step
        :param name: 如果你传入的 value 不是字典，你需要传入 value 对应的名字
        :param epoch: 前端显示需要记录 epoch
        :return:
        """
        assert isinstance(step, int) and step > -1, "Only positive integer is allowed to be `step`."
        _dict = _parse_value(value, name, parent_name='metric')
        _dict['step'] = step
        if epoch is not None:
            assert isinstance(epoch, int) and epoch > -1, "Only positive integer is allowed to be `epoch`."
            _dict['epoch'] = epoch
        _str = json.dumps(_dict)
        _str = 'Step:{}\t'.format(step) + _str
        self._write_to_logger(_str, 'metric_logger')
    
    @_check_debug
    @_check_log_dir
    def add_loss(self, value: Union[int, str, float, dict], step: int, name: str = None, epoch: int = None):
        """
        用于添加 loss。用此方法添加的值，会被记录在 loss 这一列中
        
        :param value: 类型为 int, float, str, dict中的一种。如果类型为 dict，它的键的类型只能为 str，
                它的键值的类型可以为int, float, str 或符合同样条件的 dict
        :param step: 用于和 loss 对应的 step
        :param name: 如果你传入的 value 不是字典，你需要传入 value 对应的名字
        :param epoch: 前端显示需要记录 epoch
        :return:
        """
        assert isinstance(step, int) and step > -1, "Only positive integer is allowed to be `step`."
        _dict = _parse_value(value, name, parent_name='loss')
        if epoch is not None:
            assert isinstance(epoch, int) and epoch > -1, "Only positive integer is allowed to be `epoch`."
            _dict['epoch'] = epoch
        _dict['step'] = step
        _str = json.dumps(_dict)
        _str = 'Step:{}\t'.format(step) + _str
        self._write_to_logger(_str, 'loss_logger')  # {'loss': {}, 'step':xx, 'epoch':xx}
    
    @_check_debug
    @_check_log_dir
    def add_hyper(self, value: Union[int, str, float, dict, argparse.Namespace, ConfigParser], name=None):
        """
        用于添加超参数。用此方法添加到值，会被放置在 hyper 这一列中

        :param value: 类型为 int, float, str, dict, argparse.Namespace(即ArgumentParser传入的内容), ConfigParser中的一种
                。如果类型为 dict，它的键的类型只能为 str，它的键值的类型可以为int, float, str 或符合同样条件的 dict
        :param name: 如果你传入的 value 不是字典，你需要传入 value 对应的名字
        :return:
        """
        if isinstance(value, argparse.Namespace):
            value = vars(value)
            _check_dict_value(value)
        elif isinstance(value, ConfigParser):
            value = _convert_configparser_to_dict(value)  # no need to check
        
        _dict = _parse_value(value, name=name, parent_name='hyper')
        
        self._write_to_logger(json.dumps(_dict), 'hyper_logger')
    
    @_check_debug
    @_check_log_dir
    def add_other(self, value: Union[int, str, float, dict], name: str = None):
        """
        用于添加其它参数

        :param value: 类型为 int, float, str, dict中的一种。如果类型为 dict，它的键的类型只能为 str，
                它的键值的类型可以为int, float, str 或符合同样条件的 dict
        :param name: 如果你传入 name 参数，你传入的 value 参数会被看做形如 {name:value} 的字典  
        :return:
        """
        if name in ('meta', 'hyper', 'metric', 'loss') and not isinstance(value, dict):
            raise KeyError("Don't use {} as a name. Use logger.add_{}() to _save it.".format(name, name))
        
        _dict = _parse_value(value, name=name, parent_name='other')
        self._write_to_logger(json.dumps(_dict), 'other_logger')
    
    @_check_debug
    @_check_log_dir
    def add_hyper_in_file(self, file_path: str):
        """
        从文件读取参数。如demo.py所示，两行"#######hyper"(至少5个#)之间的参数会被读取出来，并组成一个字典。每个变量最多只能出现在一行中，
        如果多次出现，只会记录第一次出现的值。demo.py::
        
            from numpy as np
            # do something
    
            ############hyper
            lr = 0.01 # some comments
            char_embed = word_embed = 300
    
            hidden_size = 100
            ....
            ############hyper
    
            # do something
            model = Model(xxx)
        
        如果你把 demo.py 的文件路径传入此函数，会转换出如下字典，并添加到参数中::
        
            {
                'lr': '0.01',
                'char_embed': '300'
                'word_embed': '300'
                'hidden_size': '100'
            }
        

        :param file_path: 文件路径
        :return:
        """
        file_path = os.path.abspath(file_path)
        if not os.path.isfile(file_path):
            raise RuntimeError("{} is not a regular file.".format(file_path))
        if not file_path.endswith('.py'):
            raise RuntimeError("{} is not a python file.".format(file_path))
        _dict = {}
        between = False
        with open(file_path, 'r', encoding='utf-8') as f:
            lines = f.readlines()
            for line in lines:
                line = line.strip()
                if len(re.findall('^#####+hyper$', line)) != 0:
                    between = not between
                elif between:
                    if len(line) != 0 and not line.startswith('#'):
                        line = re.sub(r'#[^#]*$', '', line).strip()  # 删除结尾的注释
                        # replace space before an after =
                        line = re.sub(r'\s*=\s*', '=', line)
                        values = line.split('=')
                        # 删除str开头结尾的'"
                        last_value = values[-1].rstrip("'").rstrip('"').lstrip("'").lstrip('"')
                        for value in values[:-1]:
                            _dict[value] = last_value
        if len(_dict) != 0:
            self.add_hyper(_dict)
    
    @_check_debug
    @_check_log_dir
    def add_progress(self, total_steps: int = None):
        """
        用于前端显示当前进度条。传入总的step数量
        
        :param total_steps: int, 总共有多少个step
        :return:
        """
        assert isinstance(total_steps, int) and total_steps > 0
        if hasattr(self, 'total_steps'):
            raise RuntimeError("Cannot set total_steps twice.")
        self.total_steps = total_steps
        self._write_to_logger(json.dumps({"total_steps": total_steps}), 'progress_logger')

    # @_check_debug
    # def set_rng_seed(self, rng_seed:int = None, random:bool = True, numpy:bool = True,
    #                  pytorch:bool=True, deterministic:bool=True):
    #     """
    #     设置模块的随机数种子。由于pytorch还存在cudnn导致的非deterministic的运行，所以一些情况下可能即使seed一样，结果也不一致
    #         需要在fitlog.commit()或fitlog.set_log_dir()之后运行才会记录该rng_seed到log中
    #     :param int rng_seed: 将这些模块的随机数设置到多少，默认为随机生成一个。
    #     :param bool, random: 是否将python自带的random模块的seed设置为rng_seed.
    #     :param bool, numpy: 是否将numpy的seed设置为rng_seed.
    #     :param bool, pytorch: 是否将pytorch的seed设置为rng_seed(设置torch.manual_seed和torch.cuda.manual_seed_all).
    #     :param bool, deterministic: 是否将pytorch的torch.backends.cudnn.deterministic设置为True
    #     """
    #     if rng_seed is None:
    #         import random
    #         import sys
    #         rng_seed = random.randint(sys.maxsize)
    #     if random:
    #         import random
    #         random.seed(rng_seed)
    #     if numpy:
    #         try:
    #             import numpy
    #             numpy.random.seed(rng_seed)
    #         except:
    #             pass
    #     if pytorch:
    #         try:
    #             import torch
    #             torch.manual_seed(rng_seed)
    #             torch.cuda.manual_seed_all(rng_seed)
    #             if deterministic:
    #                 torch.backends.cudnn.deterministic = True
    #         except:
    #             pass
    #     if self.initialized:
    #         self.add_other(rng_seed, 'rng_seed')

    @_check_debug
    @_check_log_dir
    def _save(self):
        if len(self._cache) != 0:
            self._create_log_files()
            for value, logger_name in self._cache:
                logger = getattr(self, logger_name)
                logger.info(value)
            self._cache = []
    
    def _write_to_logger(self, _str: str, logger_name: str):
        """
        把记录的内容写到logger里面`
        
        :param _str: 要记录的内容
        :param logger_name: 所用logger的名称
        :return:
        """
        assert isinstance(logger_name, str) and isinstance(_str, str)
        if self.save_on_first_metric_or_loss:
            if logger_name == 'metric_logger' or logger_name == 'loss_logger':
                self._create_log_files()
                self._save()  # 将之前的内容存下来
        if hasattr(self, logger_name):
            logger = getattr(self, logger_name)
            logger.info(_str.replace('\n', ' '))
        else:  # 如果还没有初始化就先cache下来
            self._cache.append([_str, logger_name])


def _convert_configparser_to_dict(config: ConfigParser) -> dict:
    """
    将ConfigParser类型的对象转成字典
    
    :param config: 代转换的对象
    :return: 转换成的字典
    """
    _dict = {}
    for section in config.sections():
        __dict = {}
        options = config.options(section)
        for option in options:
            __dict[option] = config.get(section, option)
        _dict[section] = __dict
    
    return _dict


def _parse_value(value: Union[int, str, float, dict], name: str, parent_name: str = None) -> dict:
    """
    检查传入的value是否是符合要求的。并返回dict
    
    1 如果value是基本类型，则name不为None
    2 如果value是dict类型，则保证所有value是可以转为(int, str, float)的
    
    :param value: int, float, str或者dict类型
    :param name:
    :param parent_name:
    :return:
    """
    if name is not None:
        assert isinstance(name, str), "name can only be `str` type, not {}.".format(name)
    _dict = {}
    
    if isinstance(value, (int, float, str)):
        if name == None:
            raise RuntimeError("When value is {}, you must pass `name`.".format(type(value)))
        if parent_name != None:
            _dict = {parent_name: {name: value}}
    elif isinstance(value, dict):
        _check_dict_value(value)
    if parent_name != None and name != None:
        _dict = {parent_name: {name: value}}
    elif parent_name != None:
        _dict = {parent_name: value}
    elif name != None:
        _dict = {name: value}
    else:
        _dict = value
    return _dict


def _check_dict_value(_dict: dict, prefix: str = ''):
    """
    递归检查字典中任意字段的值是否符合要求
    
    :param _dict: 被检查的字典
    :param prefix: 递归时键值的前缀
    :return:
    """
    keys = list(_dict.keys())
    for key in keys:
        value = _dict[key]
        if isinstance(value, (int, float, str)):
            continue
        elif isinstance(value, dict):
            _check_dict_value(value, prefix=prefix + ':' + key)
        elif 'torch.Tensor' in str(type(value)):
            try:
                value = value.item()
                _dict[key] = value
            except:
                raise RuntimeError("For {}. Tensor with only one element is allowed.".format(prefix + ':' + key))
        elif 'numpy.ndarray' in str(type(value)):
            total_ele = 1
            for dim in value.shape:
                total_ele *= dim
            if total_ele != 1:
                raise RuntimeError("For {}. It should only have one element.".format(prefix + ':' + key))
            _dict[key] = value.reshape(1)[0]
        else:
            raise TypeError("Only str, int, float, one element torch Tensor or numpy.ndarray"
                            " are allowed.")





"""
Usage:
    fitlog init [<name>] [--hide] [--no-git]
    fitlog revert <fit_id>  [<path>] [--id-suffix]
    fitlog -h | --help

Arguments:
    name                    Name of the fitlog project
    fit_id                  The id of the commit you want to revert
    path                    The path to revert the old commit version

Options:
    -h --help               Show this screen.
    -v --version            Show version.
    --hide                  Hide .fitconfig inside .fitlog folder
    --not-git               Not initialize with a standard git
    --id-suffix             Use commit id as the suffix of reverted folder

Examples:
    fitlog init project     Create a your project named project
    fitlog init             Init the current directory with fitlog

"""
from docopt import docopt
from fitlog.fastgit import committer


def fit_cmd(argv=None):
    if argv:
        args = docopt(__doc__, version='fitlog v1.0', argv=argv)
    else:
        args = docopt(__doc__, version='fitlog v1.0')
    if args['init']:
        name = args['<name>'] if args['<name>'] else '.'
        committer.init_project(name, hide=args["--hide"], git=not args["--no-git"])
    elif args['revert']:
        committer.revert_to_directory(args["<fit_id>"], args["<path>"], args["--id-suffix"])


if __name__ == '__main__':
    fit_cmd()


import os
import re
import json
from collections import defaultdict
from typing import List
import threading
import time


class LogReader:
    """
    用于读取日志的类
    """
    
    def __init__(self):
        """
        self._line_counter里面的内容:{save_log_dir: {filename: (line_count, last_change_time)}}
        """
        self._log_dir = None
        self._ignore_null_loss_or_metric = True  # 如果loss和metric都是null的话，则忽略
        self._line_counter = defaultdict(lambda: None)  # 记住每个log读取到的line的数量以及修改时间
    
    def set_log_dir(self, log_dir: str):
        """
        设置 log 的存放位置
        """
        if not os.path.isdir(log_dir):
            raise RuntimeError("`{}` is not a valid directory.".format(log_dir))
        empty = True
        for _dir in os.listdir(log_dir):
            if is_dirname_log_record(os.path.join(log_dir, _dir)):
                empty = False
        if empty:
            raise RuntimeError("`{}` has no valid logs.".format(log_dir))
        
        self._log_dir = log_dir
        self._line_counter.clear()  # 删除记录，使得重新读取
    
    def read_logs(self, ignore_log_names: dict = None) -> List[dict]:
        """
        从日志存放路径读取日志

        :param ignore_log_names: 如果包含在这个里面，就不会读取该log
        :return: 如果有内容或者有更新的内容，则返回一个 list，里面每个元素都是nested的dict.
            [{
                'id':
                'metric': {nested dict},
                'meta': {},
                ...
            },{
            }]
        """
        assert self._log_dir is not None, "You have to set log_dir first."
        if ignore_log_names is None:
            ignore_log_names = {}
        dirs = os.listdir(self._log_dir)
        logs = []
        for _dir in dirs:
            if _dir in ignore_log_names:
                continue
            dir_path = os.path.join(self._log_dir, _dir)
            if is_dirname_log_record(dir_path):
                _dict, file_stats = _read_save_log(dir_path, self._ignore_null_loss_or_metric,
                                                   self._line_counter[_dir])
                if len(_dict) != 0:
                    logs.append({'id': _dir, **_dict})
                    self._line_counter[_dir] = file_stats
        return logs


def _read_save_log(_save_log_dir: str, ignore_null_loss_or_metric: bool = True, file_stats: dict = None):
    """
    给定一个包含metric.log, hyper.log, meta.log以及other.log的文件夹，返回一个包含数据的dict. 如果为null则返回空字典
    不读取loss.log, 因为里面的内容对table无意义。
    
    :param _save_log_dir: 日志存放的目录， 已经最后一级了，即该目录下应该包含metric.log等了
    :param ignore_null_loss_or_metric: 是否metric为空的文件
    :param file_stats::
    
            {
                'meta.log': [current_line, last_modified_time],
                'hyper.log':[], 'metric.log':[], 'other.log':[]
            }
            
    :return:
        _dict: {'metric': {nested dict}, 'loss': {} }
        file_stats: {'meta.log': [current_line, last_modified_time],
                     'metric.log': [, ]} # 只包含有更新的文件的内容
    """
    try:
        filenames = ['meta.log', 'hyper.log', 'metric.log', 'other.log']
        if file_stats is None:
            file_stats = {}
        for filename in filenames:
            if filename not in file_stats:
                file_stats[filename] = [-1, -1]
        empty = True
        _dict = {}
        with open(os.path.join(_save_log_dir, 'metric.log'), 'r', encoding='utf-8') as f:
            for line in f:
                if len(line.strip()) != 0:
                    empty = False
                    break
        with open(os.path.join(_save_log_dir, 'loss.log'), 'r', encoding='utf-8') as f:
            for line in f:
                if len(line.strip()) != 0:
                    empty = False
                    break
        
        if empty and ignore_null_loss_or_metric:
            return _dict, file_stats
        
        for filename in filenames:
            filepath = os.path.join(_save_log_dir, filename)
            last_modified_time = os.path.getmtime(filepath)
            if file_stats[filename][1] == last_modified_time:
                continue
            file_stats[filename][1] = last_modified_time
            start_line = file_stats[filename][0]
            __dict, end_line = _read_nonstep_log_file(filepath, start_line)
            file_stats[filename][0] = end_line
            _dict = merge(_dict, __dict, use_b=False)  # 在这里，需要以文件指定顺序，保留靠前的内容的值
    except Exception as e:
        print("Exception raised when read {}".format(os.path.abspath(filepath)))
        raise e
    return _dict, file_stats


def is_log_dir_has_step(_save_log_dir: str) -> bool:
    """
    给定log_dir, 判断是否有step数据
    
    :param _save_log_dir 日志存放的目录
    :return: 是否有step数据
    """
    if not is_dirname_log_record(_save_log_dir):
        return False
    try:
        filenames = ['loss.log', 'metric.log']
        for filename in filenames:
            filepath = os.path.join(_save_log_dir, filename)
            if not os.path.exists(filepath):
                continue
            with open(filepath, 'r', encoding='utf-8') as f:
                for line in f:
                    if line.startswith('Step:'):
                        return True
    except Exception as e:
        print("Exception raised when read {}".format(os.path.abspath(filepath)))
        return False
    return False


def _read_nonstep_log_file(filepath: str, start_line: int = 0) -> (dict, int):
    """
    给定一个filepath, 读取里面非Step: 开头的line，没一行为json，使用后面的内容覆盖前面的内容
    
    :param filepath: 读取文件的路径
    :param start_line: 从哪一行开始读取
    :return: 返回一个字典(没有内容为空)和最后读取到的行号
    """
    a = {}
    with open(filepath, 'r', encoding='utf-8') as f:
        index = -1
        for index, line in enumerate(f):
            if index < start_line:
                continue
            if not line.startswith('Step:'):  # 读取非step的内容
                line = line.strip()
                try:
                    b = json.loads(line)
                except:
                    print("Corrupted json format in {}. line:{}".format(filepath, line))
                    continue
                a = merge(a, b, use_b=True)
    return a, index + 1


def merge(a: dict, b: dict, use_b: bool = True) -> dict:
    """
    将两个dict recursive合并到a中，有相同key的，根据use_b判断使用哪个值
    
    :param a: 字典 a
    :param b: 字典 b
    :param use_b: 是否使用字典 b 的值
    :return: 返回字典 a
    """
    for key in b:
        if key in a:
            if isinstance(a[key], dict) and isinstance(b[key], dict):
                merge(a[key], b[key], use_b)
            elif use_b:
                a[key] = b[key]
        else:
            a[key] = b[key]
    return a


def is_dirname_log_record(dir_path: str) -> bool:
    """
    检查dir_path是否是一个合法的log目录。合法的log目录里必须包含meta.log。
    
    :param dir_path: 被检测的路径
    :return: 是否合法
    """
    if not os.path.isdir(dir_path):
        return False
    if len(re.findall(r'log_\d+_\d+$', dir_path)) != 0:
        filenames = ['meta.log']  # 至少要有meta.log表明这个是合法的log
        for filename in filenames:
            if not os.path.exists(os.path.join(dir_path, filename)):
                return False
        return True
    else:
        return False


def is_log_record_finish(save_log_dir: str) -> bool:
    """
    检测日志的记录是否已经结束
    
    :param save_log_dir: 日志存放的目录
    :return:
    """
    if is_dirname_log_record(save_log_dir):
        with open(os.path.join(save_log_dir, 'meta.log'), 'r', encoding='utf-8') as f:
            line = ''
            for line in f:
                pass
            if len(line.strip()) != 0:
                try:
                    _d = json.loads(line)
                except:
                    return False
                if 'state' in _d['meta'] and _d['meta']['state'] == 'finish':
                    return True
    return False


class StandbyStepLogReader(threading.Thread):
    """
    用于多线程读取日志的类
    
    :param save_log_dir: 日志存放的目录
    :param uuid: 用于唯一识别 Reader 的 uuid
    :param wait_seconds:  在文件关闭后再等待{wait_seconds}秒结束进程
    :param max_no_updates: 在{max_no_updates}秒内没有更新时结束进程
    """
    
    def __init__(self, save_log_dir: str, uuid: str, wait_seconds: int = 60, max_no_updates: int = 30):
        super().__init__()
        
        self.save_log_dir = save_log_dir
        self._file_handlers = {}
        
        self.uuid = uuid
        self._last_access_time = None
        # 如果这么长时间没有读取到新的数据，就认为是不需要再读取的了
        # 如果这么长时间没有再次调用，就关掉文件
        self._wait_seconds = wait_seconds
        
        self.unfinish_lines = {}  # 防止读写冲突, key: line
        self._stop_flag = False
        self._quit = False
        self._no_update_count = 0
        self.max_no_update = max_no_updates
        
        self._last_meta_md_time = None
        self._meta_path = os.path.join(self.save_log_dir, 'meta.log')
        self._total_steps = None
    
    def _create_file_handler(self):
        """
        检查是否有未加入的handler，有则加入进来
        
        :return:
        """
        for filename in ['metric.log', 'loss.log']:
            handler_name = filename.split('.')[0]
            if handler_name in self._file_handlers:
                continue
            filepath = os.path.join(self.save_log_dir, filename)
            handler = open(filepath, 'r', encoding='utf-8')
            self._file_handlers[handler_name] = handler
    
    def _is_finish_in_meta(self) -> bool:
        """
        检查是否已经在meta中写明了finish的状态了
        
        :return: bool
        """
        
        last_meta_md_time = os.path.getmtime(self._meta_path)
        if self._last_meta_md_time is None or self._last_meta_md_time != last_meta_md_time:
            with open(self._meta_path, 'r', encoding='utf-8') as f:
                line = ''
                for line in f:
                    pass
                line = line.strip()
                if len(line) != 0:
                    try:
                        _dict = json.loads(line)['meta']
                        if 'state' in _dict and _dict['state'] in ('finish', 'error'):
                            return True
                    except:
                        pass
        self._last_meta_md_time = last_meta_md_time
        return False
    
    @staticmethod
    def read_update_single_log(filepaths: List[str], ranges: dict) -> dict:
        """
        调用这个函数，获取filepaths中满足range_min, range_max的log
        
        :param filepaths: 完整的path路径
        :param ranges: {'metric':[min, max] }
        :return: 返回值的结构如下。loss这个list是进行了step排序的
                ::
                
                    {
                    loss: [dict('step':x, key:value, 'loss':{})],
                    metric:[dict('step':x, key:value, 'metric':)]
                    }
                
        """
        updates = defaultdict(list)
        
        for filepath in filepaths:
            filename = os.path.basename(filepath).split('.')[0]
            range_min = int(ranges[filename][0])
            range_max = int(ranges[filename][1])
            
            with open(filepath, 'r', encoding='utf-8') as f:
                for line in f:
                    if not line.endswith('\n'):  # 结尾不是回车，说明没有读完
                        pass
                    else:
                        if line.startswith('Step:'):
                            step = int(line[line.index(':') + 1:line.index('\t')])
                            if range_min <= step <= range_max:
                                line = line[line.index('\t') + 1:].strip()
                            try:
                                _dict = json.loads(line)
                                updates[filename].append(_dict)
                            except:
                                pass
                if filename in updates and len(updates[filename]) != 0:  # 对step排序，保证不要出现混乱
                    updates[filename].sort(key=lambda x: x['step'])
        return updates
    
    def read_update(self, only_once: bool = False) -> dict:
        """
        调用这个函数，获取新的更新
        
        :param only_once: 是否只读取内容一次
        :return: 返回值的结构如下
                ::
        
                    {
                        loss: [dict('step':x, key:value, 'loss':{})],
                        metric:[dict('step':x, key:value, 'metric':)],
                        finish:bool(not every time),
                        total_steps:int(only the first access)
                    }
                
        """
        updates = {}
        if not self._quit:
            flag = False
            self._create_file_handler()
            updates = defaultdict(list)
            if self._last_access_time is None:
                filepath = os.path.join(self.save_log_dir, 'progress.log')
                if os.path.exists(filepath):
                    with open(filepath, 'r', encoding='utf-8') as f:
                        line = f.readline()
                        try:
                            _dict = json.loads(line.strip())
                            if 'total_steps' in _dict:
                                self._total_steps = _dict['total_steps']
                                updates['total_steps'] = _dict['total_steps']
                        except:
                            pass
                flag = True
            self._last_access_time = time.time()
            for filename, handler in self._file_handlers.items():
                for line in handler.readlines():
                    if filename in self.unfinish_lines:
                        line = self.unfinish_lines.pop(filename) + line
                    if not line.endswith('\n'):  # 结尾不是回车，说明没有读完
                        self.unfinish_lines[filename] = line
                    else:
                        if line.startswith('Step:'):
                            line = line[line.index('\t') + 1:].strip()
                            try:
                                _dict = json.loads(line)
                                updates[filename].append(_dict)
                            except:
                                pass
                if filename in updates and len(updates[filename]) != 0:  # 对step排序，保证不要出现混乱
                    updates[filename].sort(key=lambda x: x['step'])
            if not only_once:
                if len(updates) == 0:
                    self._no_update_count += 1
                else:
                    self._no_update_count = 0
                if flag:
                    self.start()
            else:  # 如果确定只读一次，则直接关闭。应该是finish了
                self._close_file_handler()
                updates['finish'] = True
        if self._quit or self._no_update_count > self.max_no_update:
            updates = {'finish': True}
        if self._is_finish_in_meta():
            updates['finish'] = True
        if 'finish' in updates:
            self._quit = True
            self.stop()
        
        return updates
    
    def _close_file_handler(self):
        for key, handler in self._file_handlers.items():
            handler.close()
        self._file_handlers.clear()
    
    def stop(self):
        """
        如果手动停止某个任务
        
        :return:
        """
        self._stop_flag = True
        self._close_file_handler()
        count = 0
        while not self._quit:
            time.sleep(1)
            if count > 3:
                raise RuntimeError("Multi-thread bug here. It should not run twice.")
            count += 1
    
    def run(self):
        """
        重载了多线程的运行函数
        
        :return:
        """
        while time.time() - self._last_access_time < self._wait_seconds and not self._stop_flag and \
                self._no_update_count < self.max_no_update:
            time.sleep(0.5)
        self._quit = True
        self._close_file_handler()






def expand_dict(prefix, _dict, connector='-', include_fields=None):
    new_dict = {}
    for key, value in _dict.items():
        if prefix != '':
            new_prefix = prefix + connector + key
        else:
            new_prefix = key
        if isinstance(value, dict):
            new_dict.update(expand_dict(new_prefix, value, connector, include_fields))
        else:
            if include_fields is None or new_prefix in include_fields:
                new_dict[new_prefix] = value
    return new_dict


def replace_nan_inf(data):
    # data: List[dict]
    if isinstance(data, list):
        for d in data:
            _replace_nan_inf(d)
    elif isinstance(data, dict):
        _replace_nan_inf(data)
    else:
        raise TypeError("Unsupported type.")
    return data

def _replace_nan_inf(d):
    for k, value in d.items():
        if isinstance(value, dict):
            _replace_nan_inf(value)
        elif isinstance(value, list):
            for d in value:
                _replace_nan_inf(d)
        elif value==float('inf'):
            d[k] = "Infinity"
        elif value==float('-inf'):
            d[k] = "-Infinity"
        elif str(value)=='nan':
            d[k] = "NaN"

def check_uuid(gold_uuid, _uuid):
    if gold_uuid==_uuid:
        return None
    else:
        return {'status': 'fail',
                'msg': "The data are out-of-date, please refresh this page. Or, you can set this page as Offline to "
                       "stop sending updates to the server."}



from ...fastlog.log_read import StandbyStepLogReader
from .utils import expand_dict

from collections import defaultdict
import re

class ChartStepLogHandler:
    def __init__(self, save_log_dir, uuid, round_to=6, max_steps=400, wait_seconds=60,
                 exclude_columns=None, max_no_updates=30):
        self.reader = StandbyStepLogReader(save_log_dir, uuid, wait_seconds, max_no_updates)

        self._save_log_dir = save_log_dir
        self.uuid = uuid
        self.max_steps = max_steps
        self.round_to = round_to
        self.path2path = {}  #{'metric':dict, 'loss':dict}

        if exclude_columns is None:
            exclude_columns = {}
        else:
            assert isinstance(exclude_columns, dict)
        self.exclude_columns = exclude_columns

    def _add_path2path(self, key, value):
        path2spath = self.path2rpath(value)
        self.path2path[key] = path2spath
        return path2spath

    def read_single_update(self, filepaths, ranges):
        steps = self.reader.read_update_single_log(filepaths, ranges)
        data = {}
        for key, values in steps.items():# key为loss, metric; value为[{'step':, epoch:, loss:{}}],
            # [{'step':, epoch:, metric:{}}]
            if key in self.path2path:
                path2path = self.path2path[key]
            else:
                path2path = self._add_path2path(key, values[0])
            expanded_values = defaultdict(list)
            for v in values:
                expand_v = {}
                real_v = v[key]
                for _key in ['step', 'epoch']:
                    if _key in v:
                        expand_v[_key] = v[_key]
                real_v.pop('step', None)
                real_v.pop('epoch', None)
                _expand_v = expand_dict('', real_v)
                for i_key, i_value in _expand_v.items():
                    if isinstance(i_value, (float, int)):
                        if i_key not in path2path:
                            path2path = self._add_path2path(key, real_v)
                        short_i_key = path2path[i_key]
                        if short_i_key in self.exclude_columns:
                            continue
                        i_value = round(i_value, self.round_to)
                        i_expand_v = expand_v.copy()
                        i_expand_v['name'] = short_i_key
                        i_expand_v['value'] = i_value
                        expanded_values[short_i_key].append(i_expand_v)

            l_expanded_values = []
            for i_key in list(expanded_values.keys()):
                i_value = expanded_values[i_key]
                l_expanded_values.extend(i_value)
            data[key] = l_expanded_values
        return data

    def update_logs(self, only_once=False):
        # data: {'finish':(如果结束了有), 'metric': [{}, {}], 'loss':[{}, {}]}
        steps = self.reader.read_update(only_once)
        data = {}
        for key, values in steps.items():# key为loss, metric, value为[{'step':, epoch:, loss:{}}]
            # [{'step':, epoch:, metric:{}}]
            if key!='finish' and key!='total_steps':
                if key in self.path2path:
                    path2path = self.path2path[key]
                else:
                    path2path = self._add_path2path(key, values[0])
                expanded_values = defaultdict(list)
                for v in values:
                    expand_v = {}
                    real_v = v[key]
                    for _key in ['step', 'epoch']:
                        if _key in v:
                            expand_v[_key] = v[_key]
                    real_v.pop('step', None)
                    real_v.pop('epoch', None)
                    _expand_v = expand_dict('', real_v)
                    for i_key, i_value in _expand_v.items():
                        if isinstance(i_value, (float, int)):
                            # TODO 可能需要精简一下路径长度， 比如BMESMetric之类的东西
                            if i_key not in path2path:
                                path2path = self._add_path2path(key, real_v)
                            short_i_key = path2path[i_key]
                            if short_i_key in self.exclude_columns:
                                continue
                            i_value = round(i_value, self.round_to)
                            i_expand_v = expand_v.copy()
                            i_expand_v['name'] = short_i_key
                            i_expand_v['value'] = i_value
                            expanded_values[short_i_key].append(i_expand_v)

                l_expanded_values = []
                for i_key in list(expanded_values.keys()):
                    i_value = expanded_values[i_key]
                    if len(i_value)>self.max_steps: # 不能超过一定的step
                        l_expanded_values.extend(i_value[-self.max_steps:])
                    else:
                        l_expanded_values.extend(i_value)
                data[key] = l_expanded_values
            else:
                data[key] = values
        return data

    def path2rpath(self, _dict):
        """
        比如两个value的path
            dev-BMESMetric-f1
            dev-BMESMEtric-pre
            test-BMESMetric-f1
        反着写(删除Metric):
            f1-BMES-dev
            pre-BMES-dev

        :param _dict: {'expanded_path': 'short_path'}
        :return:
        """
        paths = _get_dict_path(_dict)
        path2path = _reverse_path(paths)
        return path2path

def _get_dict_path(_dict, paths=None):
    # 给定一个dict, 返回所有的path，path以[[path], []]展示. 内容存到container中
    if paths==None:
        paths = []
    else:
        paths = paths.copy()
    new_paths = []
    for key, value in _dict.items():
        if isinstance(value, dict):
            _paths = _get_dict_path(value, paths + [key])
            new_paths.extend(_paths)
        else:
            new_paths.append(paths + [key])
    return new_paths

def _reverse_path(paths):
    """

    给定list的path,
        [['BMESF1Metric', 'f1'], ['BMESF1Metric'], ...]
    :param paths: {'BMESF1Metric':'f1-BMESF1Metric'}即把内容放到前面
    :return:
    """
    path2path = {}
    for path in paths:
        new_path = []
        for key in path:
            for span in re.finditer('[Mm]etric$', key):
                key = key[:span.start()]
            new_path.append(key)
        path2path['-'.join(path)] = '-'.join(reversed(new_path))
    return path2path


def _refine_path(paths):
    """
    给定list的path，将公共的部分删掉一些. 这里只处理完全一样深度的metric. 主要为了删除相同的metric_name
        [['metric', 'BMESF1MEtric', 'f1'], ['metric', 'BMESF1Metric'], ...]
    :param paths:
    :return:
    """
    if len(set(map(len, paths)))!=1:# 如果深度不同直接回去
        path2shortpath = {'-'.join(path):'-'.join(path) for path in paths}
    elif len(paths)==0:
        path2shortpath = {'-'.join(paths[0]): paths[0][-1]}
    else:
        delete_depths = []
        for depth in range(len(paths[0])):
            names = set()
            for path in paths:
                names.add(path[depth])
            if len(names)==1:
                delete_depths.append(depth)
        for i in range(len(paths)):
            for d in reversed(delete_depths):
                paths[i].pop(d)
        path2shortpath = {'-'.join(path): '-'.join(path) for path in paths}
    return path2shortpath

def _refine_logs(logs, max_points, round_to=6):
    if len(logs)<max_points:
        return logs
    log_per_bin = len(logs)//max_points
    left_log_num = len(logs) - len(logs)//max_points*max_points
    new_logs = []
    for i in range(min(max_points, len(logs))):
        _dict = defaultdict(list)
        start_idx = max(i*log_per_bin, i)
        for j in range(log_per_bin):
            k = start_idx + j
            log = logs[k]
            _dict['value'].append(log['value'])
        if left_log_num>-1:
            log = logs[start_idx + log_per_bin]
            _dict['value'].append(log['value'])
            left_log_num -= 1
        _dict['value'] = round(sum(_dict['value'])/len(_dict['value']), round_to)
        _dict['step'] = int(log_per_bin/2 + start_idx)
        _dict['name'] = log['name']
        new_logs.append(_dict)
    return new_logs


if __name__ == '__main__':
    a = {'test':{'F1SpanMetric': {'f1': 0.3606739335382445, 'pre': 0.6220669896180324}}, 'dev':{'F1SpanMetric': {'f1': 0.45963128728272284, 'pre1': 0.25299515839718356}}}

    paths = _get_dict_path(a)
    print(paths)
    print(_reverse_path(paths))

    print(expand_dict('', a))



"""
Usage:
    fitlog log <log-dir> [--log-config-name=L] [--port=P] [--standby-hours=S] [--token=T]

Arguments:
    log-dir                 Where to find logs.

Options:
    -h --help               This is a command to start fitlog server to visualize logs.
    -l=L --log-config-name  Log server config name. Must under the folder of <log-dir>. [default: default.cfg]
    -p=P --port             Which port to start to looking for usable port.[default: 5000]
    -s=S --standby-hours    How long to wait before the server . [default: 48]
    -t=T --token            If this is used, your have to specify the token when accessing. Default no token.
"""
from docopt import docopt
import os
from ..fastserver.app import start_app


def log_cmd(argv=None):
    if argv:
        args = docopt(__doc__, version='fitlog v1.0', argv=argv)
    else:
        args = docopt(__doc__, version='fitlog v1.0')

    log_dir = args['<log-dir>']
    start_port = int(args['--port'])
    log_config_name = args['--log-config-name']
    standby_hours = int(args['--standby-hours'])
    token = args['--token']
    if token is False:
        token = None
    if not os.path.isabs(log_dir):
        cwd = os.getcwd()
        log_dir = os.path.join(cwd, log_dir)

    if log_config_name!=None and not log_config_name.endswith('.cfg'):
        raise RuntimeError("log_config_name has to end with .cfg.")

    if not os.path.exists(log_dir):
        raise RuntimeError("{} is not exist.".format(log_dir))
    
    if not os.path.isdir(log_dir):
        raise NotADirectoryError("{} is not a directory.".format(log_dir))
    
    log_dir = os.path.abspath(log_dir)
    if os.path.dirname(log_config_name) != '':
        raise ValueError("log_config_name can only be a filename.")
    
    start_app(log_dir, log_config_name, start_port, standby_hours, token)


if __name__ == '__main__':
    log_cmd()





import argparse
import socket
import time
from urllib import request as urequest
import threading

def cmd_parser():
    # 返回为app.py准备的command line parser
    parser = argparse.ArgumentParser(description="To display your experiment logs in html.")

    parser.add_argument('-d', '--log_dir', help='Where to read logs. This directory should include a lot of logs.',
                         required=True, type=str)
    parser.add_argument('-l', '--log_config_name',
                        help="Log config name. Will try to find it in {log_dir}/{log_config_name}. Default is "
                             "default.cfg",
                        required=False,
                        type=str, default='default.cfg')
    parser.add_argument('-p', '--port', help='What port to use. Default 5000, but when it is blocked, pick 5001 ...',
                         required=False, type=int, default=5000)

    return parser

def get_usage_port(start_port):
    # 给定一个start_port, 依次累加直到找到一个可用的port
    while start_port<65535:
        if net_is_used(start_port):
            start_port += 1
        else:
            return start_port

def net_is_used(port, ip='0.0.0.0'):
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    try:
        s.connect((ip, port))
        s.shutdown(2)
        return True
    except:
        return False

class ServerWatcher(threading.Thread):
    def __init__(self, LEAST_REQUEST_TIMESTAMP):
        super().__init__()
        self.deque = LEAST_REQUEST_TIMESTAMP
        self._stop_flag = False

    def set_server_wait_seconds(self, server_wait_seconds):
        self.server_wait_seconds = server_wait_seconds

    def run(self):
        while (time.time() - self.deque[0])<self.server_wait_seconds and not self._stop_flag:
            time.sleep(1)
        print("This server is going to shut down.")
        try:
            if not self._stop_flag:  # 不是手动关闭的
                req = urequest.Request('http://localhost:5000/kill', headers={}, data=''.encode('utf-8'))
                page = urequest.urlopen(req).read().decode('utf-8')
        except Exception as e:
            print(e)
            raise RuntimeError("Error occurred when try to automatically shut down server.")

    def stop(self):
        self._stop_flag = True




import os
import sys
import configparser
from datetime import datetime
from fnmatch import fnmatch
import subprocess
from typing import List, Union

_commit_flag = '-------commit-------\n'
_system_flag = '-------system-------\n'
_arguments_flag = '------arguments-----\n'


class Commit(list):
    """Commit 是一个有两个元素的 list，第一个元素是commit-id，第二个元素是commit-message

    :param commit_id: 本次 commit 的 commit-id
    :param msg: 本次 commit 的 commit-message
    :return:
    """
    
    def __init__(self, commit_id: str, msg: str):
        list.__init__(self)
        self.append(commit_id)
        self.append(msg)


class Info(dict):
    """Info 是一个dict，有 status 和 msg 两个字段, 用于给返回信息加上状态码

    :param status: 状态码，0表示没有错误
    :param msg: 返回信息
    :return:
    """
    
    def __init__(self, status: int, msg: Union[str, Commit, List[str], List[Commit]]):
        dict.__init__(self)
        self["status"] = status
        self["msg"] = msg


def _colored_string(string: str, color: str or int) -> str:
    """在终端中显示一串有颜色的文字
    
    :param string: 在终端中显示的文字
    :param color: 文字的颜色
    :return:
    """
    if isinstance(color, str):
        color = {
            "black": 30,
            "red": 31,
            "green": 32,
            "yellow": 33,
            "blue": 34,
            "purple": 35,
            "cyan": 36,
            "white": 37
        }[color]
    return "\033[%dm%s\033[0m" % (color, string)


class Committer:
    """
    用于自动 commit 的类，fastgit子模块使用此类实现单例模式。实例化后的对象会记录工作目录，配置文件路径等信息。
    对外只暴露有用的接口，内部处理函数的文档请在代码中查看。
    
    """
    
    def __init__(self):
        self.work_dir = None
        self.config_file_path = None
        self.revert_with_commit_id = False  # revert后的文件夹是否以 commit_id 做为后缀
        self.watched_rules = []  # 记录被监控文件的规则
        self.commits = []  # 自动 commit 的历史记录
        self.last_commit = None  # 上一次 commit 的历史记录
    
    def _find_config_file(self, run_file_path: str = None, cli: bool = True) -> str:
        """
        
        :param run_file_path: 执行 commit 操作文件的目录
        :param cli: 是否在命令行输出信息
        :return: 返回workdir，同时在 self.config_file_path 中存储配置文件路径
        """
        config_file_name = '.fitconfig'
        if run_file_path is None:
            path = os.getcwd()
        else:
            path = os.path.abspath(run_file_path)
        home_path = os.path.expanduser('~')
        root_path = "/"
        depth_cnt = 0
        max_depth = 8
        while 1:
            depth_cnt += 1
            if max_depth == depth_cnt:
                if cli:
                    print(_colored_string("Folder depth out of limitation.", "red"))
                    print(_colored_string("Can not find the config file.", "red"))
                return "Error"
            if os.path.isfile(path + "/" + config_file_name):
                self.config_file_path = path + "/" + config_file_name
                self.work_dir = path
                return path
            if os.path.isfile(path + "/.fitlog/" + config_file_name):
                self.config_file_path = path + "/.fitlog/" + config_file_name
                self.work_dir = path
                return path
            if path == home_path or path == root_path:
                if cli:
                    print(_colored_string("Reach the root or home.", "red"))
                    print(_colored_string("Can not find the config file.", "red"))
                return "Error"
            path = os.path.dirname(path)
    
    def _read_config(self):
        """在获取配置文件路径后，读取其中的配置。采取保守策略，遇到错误自动跳过。
        
        :return:
        """
        config = configparser.ConfigParser()
        config.read(self.config_file_path)
        self.config = config
        if 'fit_settings' in config:
            if 'watched_rules' in config['fit_settings']:
                tmp = config['fit_settings']['watched_rules']
                self.watched_rules = [each.strip() for each in tmp.split(',') if len(each) > 1]
            if 'revert_with_commit_id' in config['fit_settings']:
                tmp = config['fit_settings']['revert_with_commit_id']
                if tmp in ['False', 'false']:
                    self.revert_with_commit_id = False
                elif tmp in ["True", 'true']:
                    self.revert_with_commit_id = True
                else:
                    print(
                        _colored_string("The config field 'revert_with_commit_id' (value: %s) error." % tmp, "yellow"))
    
    def _get_watched_files(self) -> List[str]:
        """在获取监管文件的规则后，获取具体的文件列表
        
        :return: 即将被 commit 的文件的列表
        """
        rules = self.watched_rules
        watched_files = []
        to_match = []
        for rule in rules:
            t = rule.count('*')
            if t == 0:
                watched_files.append(rule)
            elif t == 1:
                to_match.append(rule)
        for parent, _, file_names in os.walk(self.work_dir):
            for file_name in file_names:
                for each in to_match:
                    if fnmatch(file_name, each):
                        watched_files.append(os.path.join(parent, file_name))
                        break
        return watched_files
    
    @staticmethod
    def _switch_to_fast_git(work_dir: str) -> List[str]:
        """将工作目录从通常的 git 模式切换成 fastgit 模式
        
        :param work_dir: 工作目录的绝对路径
        :return: 返回系统输出的情况，用于出现错误时的调试
        """
        commands = ["cd " + work_dir]
        if os.path.exists(work_dir + "/.git"):
            commands.append("mv .git .git_backup")
        if os.path.isfile(work_dir + "/.gitignore"):
            commands.append("mv .gitignore .gitignore_backup")
        commands.append("mv .fitlog .git")
        commands.append("mv .git/.gitignore .")
        command = " && ".join(commands)
        return os.popen(command).readlines()
    
    @staticmethod
    def _switch_to_standard_git(work_dir: str) -> List[str]:
        """将工作目录从 fastgit 模式切换成通常的 git 模式
        
        :param work_dir: 工作目录的绝对路径
        :return: 返回系统输出的情况，用于出现错误时的调试
        """
        commands = ["cd " + work_dir, "mv .git .fitlog"]
        commands += ["mv .gitignore .fitlog/"]
        if os.path.exists(work_dir + "/.git_backup"):
            commands.append("mv .git_backup .git", )
        if os.path.isfile(work_dir + "/.gitignore_backup"):
            commands.append("mv .gitignore_backup .gitignore")
        command = " && ".join(commands)
        return os.popen(command).readlines()
    
    @staticmethod
    def _check_directory(work_dir: str, cli: bool = True) -> bool:
        """检查指定目录是否已经存在 fitlog 项目
        
        :param work_dir: 工作目录的绝对路径
        :param cli: 是否在命令行输出信息
        :return: 返回是否存在 fitlog 项目
        """
        if os.path.exists(work_dir + "/.fitlog"):
            if cli:
                print(_colored_string("Fitlog project has been initialized. ", "red"))
            return True
        return False
    
    def _commit_files(self, watched_files: List[str], commit_message: str) -> List[str]:
        """利用当前 git（如果运行正常，应该是 fastgit 模式）进行一次 commit
        
        :param watched_files: 即将被 commit 的文件的列表
        :param commit_message: commit的message
        :return: 返回系统输出的情况，用于出现错误时的调试
        """
        commands = ["cd " + self.work_dir]
        for file in watched_files:
            commands.append("git add " + file)
        commands.append("git add .gitignore")
        commands.append("git commit -m \"%s\"" % commit_message)
        command = " && ".join(commands)
        return os.popen(command).readlines()
    
    def _save_log(self, logs: List[str]):
        """ 将要存储的信息存储到默认的 fitlog
        
        :param logs: 要存储的信息
        :return:
        """
        with open(self.work_dir + "/.fitlog/fit_logs", "a")as file_out:
            file_out.writelines(logs)
    
    def _get_commits(self, cli: bool = False) -> Info:
        """从项目目录下的记录获取 fastgit 的所有 commit-id
        
        :param cli:
        :return: 返回带状态码的信息。如果成功，信息为 fastgit 的所有 commit-id 的数组
        """
        if cli:
            work_dir = os.path.abspath('.')
        else:
            if self.work_dir is None:
                return Info(1, "Error: Have not set the work directory")
            work_dir = self.work_dir
        try:
            master = work_dir + "/.fitlog/logs/refs/heads/master"
            with open(master, "r") as fin:
                lines = fin.readlines()
            commit_ids = []
            for line in lines:
                commit_ids.append(line.split()[1])
            return Info(0, commit_ids)
        except FileNotFoundError:
            return Info(1, "Error: Some error occurs")
    
    def _get_last_commit(self, cli: bool = False) -> Info:
        """从项目目录下的记录获取 fastgit 的上一次 commit-id
        
        :param cli:
        :return: 返回带状态码的信息。如果成功，信息为 fastgit 的上一次 commit-id
        """
        info = self._get_commits(cli)
        if info["status"] == 1:
            return Info(1, "Error: Not a git repository (or no commit)")
        else:
            commit_ids = info["msg"]
            if len(commit_ids) >= 1:
                return Info(0, commit_ids[-1])
            else:
                return Info(1, "Error: Not a git repository (or no commit)")
    
    def _revert(self, commit_id: str, path: str = None, cli: bool = False, id_suffix: bool = False) -> Info:
        """回退 fastgit 的一个目标版本到指定放置路径
        
        :param commit_id: 回退版本的 commit-id
        :param path: 回退版本的指定放置路径
        :param cli:
        :param id_suffix: 回退版本的放置文件夹是否包含 commit-id 做为后缀
        :return: 返回带状态码的信息。如果成功，信息为回退版本的放置路径
        """
        if cli:
            work_dir = os.path.abspath('.')
        else:
            if self.work_dir is None:
                return Info(1, "Error: Have not set the work directory")
            work_dir = self.work_dir
        if len(commit_id) < 6:
            if cli:
                print(_colored_string("Commit-id's length is at least 6", "red"))
            return Info(1, "Error: Commit-id's length is at least 6")
        if self._check_directory(work_dir, cli=False):
            commit_ids = self._get_commits()['msg']
            flag = False
            for full_commit_id in commit_ids:
                if full_commit_id.startswith(commit_id):
                    flag = True
                    commit_id = full_commit_id
                    break
            if not flag:
                if cli:
                    print(_colored_string('Can not find the commit %s' % commit_id, 'red'))
                return Info(1, 'Error: Can not find the commit %s' % commit_id)
            else:
                if path is None:
                    path = work_dir + "-revert"
                else:
                    path = os.path.abspath(path)
                if self.revert_with_commit_id or id_suffix:
                    path += "-" + commit_id[:6]
                
                if os.path.abspath(path).startswith(work_dir + '/'):
                    if cli:
                        print(_colored_string("The <path> can't in your project directory.", "red"))
                    return Info(1, "Error: The <path> can't in your project directory.")
                else:
                    ret_code = os.system("mkdir -p %s && /bin/cp -rf %s/.fitlog %s/.fitlog" % (path, work_dir, path))
                    if ret_code != 0:
                        if cli:
                            print(_colored_string("Some error occurs in cp", "red"))
                        return Info(1, "Error: Some error occurs in cp")
                self._switch_to_fast_git(path)
                ret_code = os.system("cd %s && git reset --hard %s" % (path, commit_id))
                self._switch_to_standard_git(path)
                if ret_code != 0:
                    if cli:
                        print(_colored_string("Some error occurs in git reset", "red"))
                    return Info(1, "Error: Some error occurs in git reset")
                if cli:
                    print("Your code is reverted to " + _colored_string(path, "green"))
                return Info(0, path)
        else:
            if cli:
                print(_colored_string('Not in a fitlog directory', 'red'))
            return Info(1, "Error: Not in a fitlog directory")
    
    # 对用户暴露的接口
    def commit(self, file: str, commit_message: str = None) -> Info:
        """用户用该方法进行 commit
        
        :param file: 执行文件路径，期望传入用户程序中的 __file__
        :param commit_message: 自动 commit 的 commit-message
        :return: 返回带状态码的信息。如果成功，信息为 commit-id
        """
        if commit_message is None:
            commit_message = "Commit by fitlog"
        if self.config_file_path is None:
            self._find_config_file(file)
            if self.config_file_path is None:
                return Info(1, "Error: Config file is not found")
            self._read_config()
            if not os.path.exists(self.work_dir + "/.fitlog"):
                print(_colored_string(".fitlog folder is not found", "red"))
                return Info(1, "Error: .fitlog folder is not found")
        commit_files = self._get_watched_files()
        if commit_files is None:
            return Info(1, "Error: no file matches the rules")
        logs = [_arguments_flag, "Run ", " ".join(sys.argv), "\n"]
        logs += [_system_flag]
        logs += self._switch_to_fast_git(self.work_dir)
        commit_files = self._get_watched_files()
        msg = self._commit_files(commit_files, commit_message)
        logs += msg
        if msg:
            print(_colored_string('Auto commit by fitlog', 'blue'))
        logs += self._switch_to_standard_git(self.work_dir)
        commit_id = self._get_last_commit()['msg']
        self.last_commit = Commit(commit_id, commit_message)
        self.commits.append(self.last_commit)
        logs = [commit_id, '\n'] + logs
        logs = [datetime.now().strftime('%Y-%m-%d %H:%M:%S'), "\n"] + logs
        logs = [_commit_flag] + logs
        self._save_log(logs + ['\n\n'])
        return Info(0, commit_id)
    
    # 对包内组件暴露的接口
    def fitlog_last_commit(self) -> Commit:
        """返回 self.last_commit 中记录的上一次的commit
        
        :return: Commit是一个元组，第一个参数为 commit-id，第二个参数为 commit-message
        """
        return self.last_commit
    
    def fitlog_commits(self) -> List[Commit]:
        """返回 self.commits 中记录的所有的commit
        
        :return: 返回一个 Commit 类型的数组
        """
        return self.commits
    
    def get_config(self, run_file_path: str = None) -> Info:
        """通过执行文件的路径获取配置信息
        
        :param run_file_path: 执行文件的路径
        :return: 返回带状态码的信息。如果成功，信息为工作目录的路径
        """
        self._find_config_file(run_file_path, cli=False)
        if self.config_file_path is None:
            return Info(1, "Error: Config file is not found")
        self._read_config()
        return Info(0, self.work_dir)
    
    def fitlog_revert(self, commit_id: str, run_file_path: str = None, id_suffix: bool = False) -> Info:
        """fitlog 调用此接口进行版本回退
        
        :param commit_id: 需要回退版本的 commit-id
        :param run_file_path: 执行文件的路径 TODO:检查这个变量是否需要
        :param id_suffix: 回退版本的放置文件夹是否包含 commit-id 做为后缀
        :return: 返回带状态码的信息。如果成功，信息为回退版本的放置路径
        """
        if self.work_dir is None:
            info = self.get_config(run_file_path)
            if info['status'] == 1:
                return info
        return self._revert(commit_id, id_suffix=id_suffix, cli=False)
    
    @staticmethod
    def git_last_commit(work_dir: str) -> Info:
        """获取标准 git 的上一次 commit 的信息

        :param work_dir: 工作目录的路径
        :return: 返回带状态码的信息。如果成功，信息为一个 Commit 类型的 commit 信息
        """
        work_dir = os.path.abspath(work_dir)
        try:
            # 忽略错误信息
            lines = subprocess.Popen("cd %s && git log --oneline" % work_dir, shell=True,
                                     stdout=subprocess.PIPE, stderr=subprocess.DEVNULL).stdout.readlines()
            if len(lines) != 0:
                line = lines[0].decode('utf-8')
                git_id = line[:line.index(' ')]
                git_msg = line[line.index(' ') + 1:].strip()
            else:
                git_id = None
                git_msg = None
            return Info(0, Commit(git_id, git_msg))
        except FileNotFoundError:
            return Info(1, "Error: Some error occurs")
    
    @staticmethod
    def fit_last_commit(work_dir: str) -> Info:
        """获取标准 fit 的上一次 commit 的信息

        :param work_dir: 工作目录的路径
        :return: 返回带状态码的信息。如果成功，信息为一个 Commit 类型的 commit 信息
        """
        work_dir = os.path.abspath(work_dir)
        try:
            Committer._switch_to_fast_git(work_dir)
            # 忽略错误信息
            lines = subprocess.Popen("cd %s && git log --oneline" % work_dir, shell=True,
                                     stdout=subprocess.PIPE, stderr=subprocess.DEVNULL).stdout.readlines()
            if len(lines) != 0:
                line = lines[0].decode('utf-8')
                git_id = line[:line.index(' ')]
                git_msg = line[line.index(' ') + 1:].strip()
            else:
                git_id = None
                git_msg = None
            
            Committer._switch_to_standard_git(work_dir)
            return Info(0, Commit(git_id, git_msg))
        except FileNotFoundError:
            return Info(1, "Error: Some error occurs")
    
    # 命令行操作所需的组件
    def revert_to_directory(self, commit_id: str, path: str, id_suffix: bool):
        """命令行用来回退 fastgit 的一个目标版本到指定放置路径
        
        :param commit_id: 回退版本的 commit-id
        :param path: 回退版本的指定放置路径
        :param id_suffix: 回退版本的放置文件夹是否包含 commit-id 做为后缀
        :return:
        """
        self._revert(commit_id, path, cli=True, id_suffix=id_suffix)
    
    def init_project(self, pj_name: str, version: str = "normal", hide: bool = False, git: bool = True) -> int:
        """命令行用来初始化一个 fitlog 项目
        
        :param pj_name: 项目名称
        :param version: 项目初始化文件夹的版本，目前只有 normal 这一种
        :param hide: 是否隐藏 .fitconfig 文件到 .fitlog 文件夹中
        :param git: 是否在初始化 fitlog 的同时为项目初始化 git
        :return: 状态码。0表示正常完成，其它错误码与系统相关
        """
        if pj_name == '.':
            if os.path.exists(".fitlog"):
                print(_colored_string("Fitlog project existed.", "red"))
                return 0
            if os.path.exists(".git"):
                self._switch_to_fast_git(os.path.abspath(pj_name))
        elif os.path.exists(pj_name + "/.fitlog"):
            print(_colored_string("Fitlog project %s existed." % pj_name, "red"))
            return 0
        tools_path = os.path.realpath(__file__)[:-len("committer.py")]
        commands = [
            "cd " + pj_name,
            "cp -r %s/. ." % (tools_path + version),
            "mv main main.py",
            "git init"
        ]
        if pj_name != '.':
            commands = ["mkdir " + pj_name] + commands
        if hide:
            commands.append("mv .fitconfig .git")
        ret_code = os.system(" && ".join(commands))
        if ret_code != 0:
            print(_colored_string("Some error occurs.", "red"))
            return ret_code
        self._switch_to_standard_git(os.path.abspath(pj_name))
        
        self.commit(pj_name + "/main.py", "Project initialized.")
        
        if git:
            if pj_name == '.' and os.path.exists(".git"):
                if hide:
                    open('.gitignore', 'a').write(".fitlog\nlogs\n.gitignore\n")
                else:
                    open('.gitignore', 'a').write(".fitlog\n.fitconfig\nlogs\n.gitignore\n")
            else:
                commands = [
                    "cd " + pj_name,
                    "git init"]
                if hide:
                    commands += ["echo .fitlog > .gitignore"]
                else:
                    commands += ["echo \".gitignore\\n.fitlog/\\n.fitconfig\\nlogs/\" > .gitignore"]
                ret_code = os.system(" && ".join(commands))
                if ret_code != 0:
                    print(_colored_string("Some error occurs.", "red"))
                    return ret_code
        print(_colored_string("Fitlog project %s is initialized." % pj_name, "green"))
        return 0
    
    def short_logs(self, show_now: bool = False, last_num: int = None):
        """在命令行用来查看 fastgit 的自带 logs
        
        :param show_now: 是否显示当前版本在 logs 中的位置
        :param last_num: 显示最近的 {last_num} 条记录
        :return:
        """
        head_id, log = None, None
        if self._check_directory(os.path.abspath('.'), cli=False):
            try:
                if show_now:
                    work_dir = os.path.abspath('.')
                    head_id = self._get_last_commit(work_dir)["msg"]
                with open('.fitlog/fit_logs', 'r') as fin:
                    lines = fin.readlines()
                cnt = 0
                show_logs = []
                for line in lines:
                    if line == _commit_flag:
                        cnt = 0
                    elif cnt == 1:
                        log = ["date&time   " + line]
                    elif cnt == 2:
                        if show_now and line.strip() == head_id:
                            log.append("commit_id   " + _colored_string(line, "green"))
                        else:
                            log.append("commit_id   " + line)
                    elif cnt == 4:
                        log.append("arguments   " + line)
                        show_logs.append(log)
                    cnt += 1
                if last_num is not None:
                    try:
                        if int(last_num) < len(show_logs):
                            show_logs = show_logs[-int(last_num):]
                    except ValueError:
                        print(_colored_string("<last_num> must be an integer.", 'red'))
                show = []
                for log in show_logs:
                    for each in log:
                        show.append(each)
                    show.append("\n")
                print("".join(show))
                if show_now:
                    print(_colored_string('Head is ' + head_id, "green"), "\n")
            except FileNotFoundError:
                print("No fitlog records here.")
        else:
            print(_colored_string('Not in a fitlog directory', 'red'))



"""
fitlog是一款集成了自动版本管理和自动日志记录两种功能的 Python 包，它可以帮助你在进行实验时方便地保存当前的代码、参数和结果。
fitlog提供给用户的 API 有如下几个：

"""
__all__ = ["commit", "set_log_dir", "finish", "add_best_metric", "add_metric", "add_loss", "add_hyper", "add_other",
           "add_hyper_in_file", "get_commit_id", "get_fit_id"]
from .fastlog import logger as _logger
from .fastgit import Committer, committer as _committer
from typing import Union
import argparse
from configparser import ConfigParser

__version__ = '0.1.2'


def get_commit_id(file):
    """用户用此命令获取上一次 Git 记录的 id, 期望的使用方法如下::
        
        id = fitlog.get_commit_id(__file__)
        
    :param file: 以该路径往上寻找.fitlog所在文件夹。一般传入__file__即可:
    :return: Git 的上次记录的 commit-id 的前七位；错误时返回 `error`
    """
    work_dir = _committer._find_config_file(file)
    res = Committer.git_last_commit(work_dir)
    if res['status'] == 0:
        return res['msg'][0]
    else:
        return 'error'


def get_fit_id(file):
    """用户用此命令获取上一次 fitlog 自动记录 commit 的 id, 期望的使用方法如下::
        
        id = fitlog.get_fit_id(__file__)
        
    
    :param file: 以该路径往上寻找.fitlog所在文件夹。一般传入__file__即可
    :return: Fitlog 的上次自动记录的 commit-id 的前七位；错误时返回 `error`
    """
    work_dir = _committer._find_config_file(file)
    res = Committer.fit_last_commit(work_dir)
    if res['status'] == 0:
        return res['msg'][0]
    else:
        return 'error'


def commit(file: str, fit_msg: str = None):
    """
    用户用此命令进行自动 commit, 期望的使用方法如下::
        
        import fitlog
        
        fitlog.commit(__file__)
        \"\"\"
        Your training code
        \"\"\"
        fitlog.finish()

    :param file: 以该路径往上寻找.fitlog所在文件夹。一般传入__file__即可
    :param fit_msg: 针对该实验的说明
    """
    _logger.commit(file, fit_msg)


def set_log_dir(log_dir: str, new_log:bool=False):
    """
    设定log 文件夹的路径(在进行其它操作前必须先指定日志路径)。如果你已经顺利执行了 fitlog.commit()命令，
    log 文件夹会自动设定为.fitconfig 文件中的 default_log_dir 字段的值。在某些情况下，可能需要继续往同
    一个log中写入数据(比如继续训练之前以及保存的模型)，可以通过将log_dir设置为具体的log名。但需要保证
    step的顺序与之前已有的内容是不冲突的，因为相同的step在fitlog中是覆盖的。

    Example::

        # 假设当前的文件结构为
        # logs/
        #    log_20190417_140311
        #    ...
        # main.py
        #以下是main.py中三种设置log位置的方式
        fitlog.commit() # 如果commit成功，则不需要设置logs文件夹了
        fitlog.set_log_dir('logs/') # 设置log文件夹为'logs/', fitlog在每次运行的时候会默认以时间戳的方式在里面生成新的log
        fitlog.set_log_dir('logs/log_20190417_140311') # fitlog将log继续写入到log_20190417_140311里。

    :param log_dir: log 文件夹的路径
    :param new_log: 是否重新创建一个log，仅在同一次python启动但是需要记录多个log时使用(但是只能分阶段地用，即同一时间
        只会有一个logger存在，设置new_log为True时，仅仅是开了一个新的logger，但同时前一个就关闭了。)
    """
    _logger.set_log_dir(log_dir, new_log)


def debug():
    """
    调用该方法之后，所有的fitlog方法都不会产生任何作用。可用于调试代码时避免输出大量无用的信息。

    Example::

        fitlog.debug()
        fitlog.commit()
        fitlog.add_metric(0.3, f1)

    由于有fitlog.debug(), commit()和add_metric()都不会实际执行的。

    :return:
    """
    _logger.debug()


def finish(status: int = 0):
    """
        使用此方法告知 fitlog 你的实验已经正确结束。你可以使用此方法来筛选出失败的实验。

        :param int status: 告知当前实验的状态。0: 结束了; 1: 发生了错误
    """
    _logger.finish(status)


def add_metric(value: Union[int, str, float, dict], step: int, name: str = None, epoch: int = None):
    """
    用于添加 metric 。用此方法添加的值不会显示在表格中，但可以在单次训练的详情曲线图中查看。

    :param value: 类型为 int, float, str, dict中的一种。如果类型为 dict，它的键的类型只能为 str，
            它的键值的类型可以为int, float, str 或符合同样条件的 dict
    :param step: 用于和 loss 对应的 step
    :param name: 如果你传入 name 参数，你传入的 value 参数会被看做形如 {name:value} 的字典
    :param epoch: 前端显示需要记录 epoch
    :return:
    """
    _logger.add_metric(value, step, name, epoch)


def add_loss(value: Union[int, str, float, dict], step: int, name: str = None, epoch: int = None):
    """
    用于添加 loss。用此方法添加的值不会显示在表格中，但可以在单次训练的详情曲线图中查看。

    :param value: 类型为 int, float, str, dict中的一种。如果类型为 dict，它的键的类型只能为 str，
            它的键值的类型可以为int, float, str 或符合同样条件的 dict
    :param step: 用于和 loss 对应的 step
    :param name: 如果你传入 name 参数，你传入的 value 参数会被看做形如 {name:value} 的字典
    :param epoch: 前端显示需要记录 epoch
    :return:
    """
    _logger.add_loss(value, step, name, epoch)


def add_best_metric(value: Union[int, str, float, dict], name: str = None):
    """
    用于添加最好的 metric 。用此方法添加的值，会被显示在表格中的 metric 列及其子列中。相同key的内容将只保留最后一次传入的值。

    :param value: 类型为 int, float, str, dict中的一种。如果类型为 dict，它的键的类型只能为 str，
            它的键值的类型可以为int, float, str 或符合同样条件的 dict
    :param name: 如果你传入 name 参数，你传入的 value 参数会被看做形如 {name:value} 的字典

    .. warning ::
        如果你在同时记录多个数据集上的performance, 请注意使用不同的名称进行区分

    """
    _logger.add_best_metric(value, name)


def add_hyper(value: Union[int, str, float, dict, argparse.Namespace, ConfigParser], name=None):
    """
    用于添加超参数。用此方法添加到值，会被放置在表格中的 hyper 列及其子列中

    :param value: 类型为 int, float, str, dict, argparse.Namespace(即ArgumentParser传入的内容), ConfigParser中的一种
            。如果类型为 dict，它的键的类型只能为 str，它的键值的类型可以为int, float, str 或符合同样条件的 dict
    :param name: 如果你传入 name 参数，你传入的 value 参数会被看做形如 {name:value} 的字典
    :return:
    """
    _logger.add_hyper(value, name)


def add_hyper_in_file(file_path: str):
    """
    从文件读取参数。如下面的文件所示，两行"#####hyper"(至少5个#)之间的参数会被读取出来，并组成一个字典。每个变量最多只能出现在一行中，
    如果多次出现，只会记录第一次出现的值。demo.py::
    
        from numpy as np
        import fitlog
        # do something

        fitlog.add_hyper_in_file(__file__)  # 会把本python文件的hyper加入进去
        ############hyper
        lr = 0.01 # some comments
        char_embed = word_embed = 300

        hidden_size = 100
        ....
        ############hyper

        # do something
        model = Model(xxx)
    
    如果你把 demo.py 的文件路径传入此函数，会转换出如下字典，并添加到参数中::
    
        {
            'lr': '0.01',
            'char_embed': '300'
            'word_embed': '300'
            'hidden_size': '100'
        }

    :param file_path: 文件路径。如果是读取本python文件中的hyper parameter可以直接fitlog.add_hyper_in_file(__file__)
    """
    _logger.add_hyper_in_file(file_path)


def add_other(value: Union[int, str, float, dict], name: str = None):
    """
    用于添加其它参数。用此方法添加到值，会被放置在表格中的 other 列及其子列中。相同key的内容将只保留最后一次传入的值。

    :param value: 类型为 int, float, str, dict中的一种。如果类型为 dict，它的键的类型只能为 str，
            它的键值的类型可以为int, float, str 或符合同样条件的 dict
    :param name: 如果你传入 name 参数，你传入的 value 参数会被看做形如 {name:value} 的字典
    """
    _logger.add_other(value, name)


def add_progress(total_steps: int = None):
    """
    传入总的step数量，用于前端计算进度。

    :param total_steps: int, 总共有多少个step
    """
    _logger.add_progress(total_steps)

# TODO 好像不work
# def set_rng_seed(rng_seed:int = None, random:bool = True, numpy:bool = True,
#                      pytorch:bool=True, deterministic:bool=True):
#     """
#     设置模块的随机数种子。由于pytorch还存在cudnn导致的非deterministic的运行，所以一些情况下可能即使seed一样，结果也不一致
#         需要在fitlog.commit()或fitlog.set_log_dir()之后运行才会记录该rng_seed到log中
#     :param int rng_seed: 将这些模块的随机数设置到多少，默认为随机生成一个。
#     :param bool, random: 是否将python自带的random模块的seed设置为rng_seed.
#     :param bool, numpy: 是否将numpy的seed设置为rng_seed.
#     :param bool, pytorch: 是否将pytorch的seed设置为rng_seed(设置torch.manual_seed和torch.cuda.manual_seed_all).
#     :param bool, deterministic: 是否将pytorch的torch.backends.cudnn.deterministic设置为True。如果该值不为True，有时候即使
#         全部随机数种子都一样也不能跑出相同的结果; 关掉的话可能会有一点性能损失。
#     """
#     _logger.set_rng_seed(rng_seed, random, numpy, pytorch, deterministic)



"""
Usage:
    fitlog <command> [<args>...]
    fitlog -h|--help
    fitlog --version

Supported commands
    init            Initialize a fitlog project
    revert          Revert to a specific version
    log             Visualize logs by a server.
    
See "fitlog help <command>" for more information on a specific command
"""
import sys
from docopt import docopt
from .fit_cmd import fit_cmd
from .log_cmd import log_cmd


def main_cmd():
    argv = sys.argv[1:2] if len(sys.argv) > 2 else sys.argv[1:]
    args = docopt(__doc__, version='fitlog v1.0', argv=argv, help=False)
    argv = sys.argv[1:]
    if args['<command>'] in ('init', 'revert'):
        if args['--help'] or args['-h']:
            fit_cmd(['-h'])
        else:
            fit_cmd(argv)
    elif args['<command>'] == 'log':
        if args['--help'] or args['-h']:
            log_cmd(['-h'])
        else:
            log_cmd(argv)
    elif args['<command>'] in ['help', None]:
        if len(argv) > 1:
            cmd = argv[1]
            if cmd == 'log':
                log_cmd(['-h'])
            elif cmd in ('init', 'revert'):
                fit_cmd(['-h'])
            else:
                print("Unknown command `{}`, only support [log, init, revert].".format(cmd))
                print(__doc__)
        else:
            print("You have to specify a command, support [log, init, revert].")
            print(__doc__)
    else:
        print("Unknown command: {}.".format(args['<command>']))
        print(__doc__)



# encoding=utf-8

# 验证如何将dict或者json文件转换为columns

import os
import json
from collections import defaultdict
from functools import reduce
from .server_config import read_extra_data
from .server_config import read_server_config
from .utils import expand_dict
from .server_config import save_config
from .server_config import save_extra_data
import warnings

def generate_columns(logs, hidden_columns=None, column_order=None, editable_columns=None,
                     exclude_columns=None, filter_condition=None, ignore_unchanged_columns=True,
                     str_max_length=20, round_to=6, ignore_not_exist=False):
    """

    :param logs: list of dict对象返回List形式的column数据.
    :param hidden_columns: {}, can chooose parent columns, then all children will be hidden
    :param column_order: dict, column的顺序
    :param editable_columns: dict，那些column是可以编辑的
    :param filter_condition: dict, 每个key的筛选条件
    :param ignore_unchanged_columns: 是否忽略不变的column
    :param int str_max_length: 长于这个的str会被以...替代
    :param round_to: int，保留多少位小数
    :param ignore_not_exist: bool, 如果不存在filter_condition中的内容，则认为是不满足条件

    return:
        data: List[]数据
        unchange_columns: dict
        column_order: dict, 前端显示的column_order
        hidden_columns: dict, 需要隐藏的column，具体到field的
        column_dict: {}只有一级内容，根据prefix取出column内容，使用DFS对order_dict访问，然后将内容增加到columns中
    """
    assert len(logs)!=0, "Empty list is not allowed."

    connector = '-'  # 不能使用“!"#$%&'()*+,./:;<=>?@[\]^`{|}~ ”
    unselectable_columns = {} # 这个dict中的prefix没有filter选项，(1) 每一行都不一样的column; (2) 只有一种value的column
    if editable_columns is None:
        editable_columns = {'memo': 1} # 在该dict的内容是editable的, 保证不会被删除
    else:
        assert isinstance(editable_columns, dict), "Only dict is allowed for editable_columns."
        editable_columns['memo'] = 1
    if exclude_columns is None:
        exclude_columns = {}
    else:
        assert isinstance(exclude_columns, dict), "Only dict is allowed for exclude_columns."
    if filter_condition is None:
        filter_condition = {}
    else:
        assert isinstance(exclude_columns, dict), "Only dict is allowed for filter_condition."

    assert isinstance(logs, list), "Only list type supported."
    for _dict in logs:
        assert isinstance(_dict, dict), "Only dict supported."
        for key,value in _dict.items():
            assert key == 'id', "`id` must be put in the first key."
            break
    if hidden_columns is not None:
        assert isinstance(hidden_columns, dict), "Only dict type suppported."
    else:
        hidden_columns = {}
    if column_order is not None:
        assert isinstance(column_order, dict), "Only dict tyep supported."
    else:
        column_order = dict()

    def add_field(prefix, key, value, fields, connector, depth):
        if prefix != '':
            prefix = prefix + connector + key
        else:
            prefix = key
        if prefix in exclude_columns: # 排除的话就不要了
            return 0
        max_depth = depth
        if isinstance(value, dict):
            for k, v in value.items():
                max_depth = max(add_field(prefix, k, v, fields, connector, depth), max_depth)
        else:
            if isinstance(value, float) and round_to is not None:
                value = round(value, round_to)
            if isinstance(value, str):
                if len(value)>str_max_length and prefix not in editable_columns: # editable的不限制
                    value = value[:str_max_length] + '...'
                    unselectable_columns[prefix] = 1
            fields[prefix] = value
        return max_depth + 1

    data = []
    max_depth = 1
    field_values = defaultdict(list) # 每种key的不同value
    exclude_log_ids = {}
    for ID, _dict in enumerate(logs):
        fields = {}
        for key, value in _dict.items():
            max_depth = max(add_field('', key, value, fields, connector, 0), max_depth)
        for key, value in fields.items():
            field_values[key].append(value)
        # field是一个一维的dict，内含expanded的key以及它的value
        filter = False
        # 如果不满足filter的条件(value不同，包含对应的key)，就不要把它添加到数据中去了
        for key, value in filter_condition.items():
            if key in fields:
                if str(value) != str(fields[key]):
                    filter = True
                    break
            elif ignore_not_exist:
                filter = True
                break
        if not filter:
            data.append(fields)
        else:
            exclude_log_ids[ID] = 1
    # 删除不满足条件的filter，如果没有满足条件的，则filter失效
    filtered_logs = [log for idx, log in enumerate(logs) if idx not in exclude_log_ids]
    if len(filtered_logs)==0:
        warnings.warn("No log meets the condition.")
    else:
        logs = filtered_logs

    # 删掉的log需要排除
    field_values = defaultdict(list) # 每种key的不同value
    for ID, _dict in enumerate(logs):
        fields = {}
        for key, value in _dict.items():
            max_depth = max(add_field('', key, value, fields, connector, 0), max_depth)
        for key, value in fields.items():
            field_values[key].append(value)

    unchange_columns = {}
    if ignore_unchanged_columns and len(logs)>1:
        must_include_columns = ['meta-fit_id', 'meta-git_id']
        for key, value in field_values.items():
            if len(value)==len(logs):
                value_set = set(value)
                # 每次都是一样的结果, 但排除只有一个元素的value以及可修改的column
                if len(value_set) == 1 and len(value)!=1 and key not in editable_columns:
                    unchange_columns[key] = value[0]  # 防止加入column中
        exclude_columns.update(unchange_columns) # 所有不变的column都不选择了
        for column in must_include_columns:
            if column in exclude_columns:
                exclude_columns.pop(column)
        for fields in data:
            for key, value in exclude_columns.items():
                if key in fields:
                    fields.pop(key)

    for key, value in field_values.items():
        value_set = set(value)
        if len(value_set) == len(value) or len(value_set)==1:
            unselectable_columns[key] = 1

    column_dict = {}  # 这个dict用于存储结构，用于创建columns. 因为需要保证创建的顺序不能乱。 Nested dict
    reduce(merge, [column_dict] + logs)
    remove_exclude(column_dict, exclude_columns)

    column_dict['memo'] = '-' # 这一步是为了使得memo默认在最后一行
    for _dict in data:
        if 'memo' not in _dict:
            _dict['memo'] = 'Click to edit'

    # 需要生成column_order
    new_column_order = dict()
    new_column_dict = {}

    # generate_columns
    new_hidden_columns = {}
    column_keys = [key for key in column_dict.keys()]

    first_column_keys = []
    for key, order_value in column_order.items(): # 先按照order_dict进行排列
        if key in column_dict:
            value = column_dict[key]
            col_span, c_col_ord = add_columns('', key, value, 0, max_depth, new_column_dict, order_value, connector,
                                              exclude_columns,
                        unselectable_columns,
                        editable_columns, hidden_columns, new_hidden_columns, False)
            column_keys.pop(column_keys.index(key))
            if col_span!=0:
                new_column_order[key] = c_col_ord
                first_column_keys.append(key)
    for key in column_keys: # 再按照剩下的排
        value = column_dict[key]
        col_span, c_col_ord = add_columns('', key, value, 0, max_depth, new_column_dict, None, connector, exclude_columns,
                    unselectable_columns,
                    editable_columns, hidden_columns, new_hidden_columns, False)
        if col_span != 0:
            new_column_order[key] = c_col_ord
            first_column_keys.append(key)
    new_column_order['OrderKeys'] = first_column_keys # 使用一个OrderKeys保存没一层的key的顺序

    data = {log['id']:log for log in data}
    res = {'data': data, 'unchanged_columns':unchange_columns, 'column_order': new_column_order, 'column_dict':new_column_dict,
           'hidden_columns': new_hidden_columns}

    return res

def remove_exclude(column_dict, exclude_dict, prefix=''):
    # 将exclude_dict中的field从nested的column_dict中移除
    keys = list(column_dict.keys())
    for key in keys:
        if prefix=='':
            new_prefix = key
        else:
            new_prefix = prefix + '-' + key
        if new_prefix in exclude_dict:
            column_dict.pop(key)
        else:
            if isinstance(column_dict[key], dict):
                remove_exclude(column_dict[key], exclude_dict, new_prefix)


def add_columns(prefix, key, value, depth, max_depth, column_dict, order_value, connector, exclude_columns,
                unselectable_columns, editable_columns, hidden_columns, new_hidden_columns, hide):
    if prefix != '':
        prefix = prefix + connector + key
    else:
        prefix = key
    if prefix in exclude_columns:
        return 0, None
    if not hide:
        if prefix in hidden_columns:
            hide = True

    item = {}
    item['title'] = key
    colspan = 1
    min_unuse_depth = max_depth - depth - 1
    column_order = {}
    if isinstance(value, dict):
        total_colspans = 0
        column_keys = [key for key in value.keys()]
        first_column_keys = []
        if isinstance(order_value, dict):# 如果order_value是dict，说明下面还有顺序
            for o_key, o_v in order_value.items():
                if o_key in value:
                    n_val = value[o_key]
                    colspan, c_col_ord = add_columns(prefix, o_key, n_val, depth + 1, max_depth, column_dict, o_v,
                                                     connector, exclude_columns, unselectable_columns, editable_columns,
                                                     hidden_columns, new_hidden_columns, hide)
                    total_colspans += colspan
                    min_unuse_depth = 0
                    column_keys.pop(column_keys.index(o_key))
                    if colspan!=0: # 需要排除这个column
                        column_order[o_key] = c_col_ord
                        first_column_keys.append(o_key)
        for key in column_keys:
            v = value[key]
            colspan, c_col_ord = add_columns(prefix, key, v, depth + 1, max_depth, column_dict, None, connector,
                                             exclude_columns, unselectable_columns, editable_columns,
                                             hidden_columns, new_hidden_columns, hide)
            total_colspans += colspan
            min_unuse_depth = 0
            if colspan != 0:  # 需要排除这个column
                column_order[key] = c_col_ord
                first_column_keys.append(key)
        colspan = total_colspans
        column_order['OrderKeys'] = first_column_keys
    else:
        item['field'] = prefix
        item['sortable'] = 'true'
        if prefix not in unselectable_columns:
            if prefix in editable_columns:
                item['filterControl'] = 'input'
            else:
                item['filterControl'] = 'select'
                item['filterStrictSearch'] = True
        if prefix in editable_columns:
            item['editable'] = 'true'
        else:
            item['editable'] = 'false'
        if hide:
            new_hidden_columns[prefix] = 1
        column_order = "EndOfOrder"

    item['rowspan'] = min_unuse_depth + 1
    item['colspan'] = colspan
    column_dict[prefix] = item

    return colspan, column_order

def merge(a, b, path=None, use_b=True):
    "merges b into a"
    # 将两个dict recursive合并到a中，有相同key的，根据use_b判断使用哪个值
    if path is None: path = []
    for key in b:
        if key in a:
            if isinstance(a[key], dict) and isinstance(b[key], dict):
                merge(a[key], b[key], path + [str(key)])
            elif use_b:
                a[key] = b[key]
        else:
            a[key] = b[key]
    return a

def prepare_incremental_data(logs, new_logs, field_columns, filter_condition=None, ignore_not_exist=False):
    """

    :param logs: {'id':dict, ...}, 之前的数据, flatten, dict.
    :param new_logs: List[dict,], 新读取到的数据, nested
    :param field_columns: {field:1}, 在前端显示的内容field
    :param filter_condition: {}, 用于过滤不需要的内容
    :param ignore_not_exist: bool, 是否删除过滤条件不存在的log
    :return: {'new_logs':[dict(), dict()...], 'update_logs':[dict(), dict()...]}
    """
    if filter_condition is None:
        filter_condition = {}

    # 1. 将new_logs的内容展平
    new_dict = {}
    for log in new_logs:
        ex_dict = expand_dict('', log, connector='-', include_fields=field_columns)
        filter = False  # 是否忽略掉
        for f_k, f_v in filter_condition.items():
            if f_k in ex_dict:
                if str(ex_dict[f_k]) != f_v:
                    filter = True
            elif ignore_not_exist:
                filter = True
        if not filter:
            new_dict[log['id']] = ex_dict

    # 2. 将logs中的内容进行替换，或增加.
    updated_logs = []
    keys = list(new_dict.keys())
    for key in keys:
        if key in logs:
            value = new_dict.pop(key)
            log = merge(logs[key], value, use_b=True)
            updated_logs.append(log)
        else:
            new_dict[key]['memo'] = 'Click to edit'
    new_logs = list(new_dict.values())

    for key, value in new_dict.items():
        logs[key] = value

    return new_logs, updated_logs


def prepare_data(log_reader, log_dir, log_config_name, all_data=None): # 准备好需要的数据， 应该包含从log dir中读取数据
    """

    :param log_reader: 用于读取数据的Reader对象
    :param log_dir: str, 哪里是存放所有log的大目录
    :param log_config_path: 从哪里读取config
    :param all_data: dict, 如果不为None则不会从硬盘读取config和extra_data

    :return:
    """
    print("Start preparing data.")
    # 1. 从log读取数据
    log_dir = os.path.abspath(log_dir)
    log_config_path = os.path.join(log_dir, log_config_name)
    log_config_path = os.path.abspath(log_config_path)

    # 读取config文件
    # 读取log_setting_path
    if all_data is None:
        all_data = {}
    all_data.update(read_server_config(log_config_path))

    deleted_rows = all_data['deleted_rows']

    logs = log_reader.read_logs(deleted_rows)

    if len(logs)==0:
        raise ValueError("No valid log found in {}.".format(log_dir))

    # read extra_data
    if 'extra_data' not in all_data: # 只有在第一次需要读取
        extra_data_path = os.path.join(log_dir, 'log_extra_data.txt')
        extra_data = {}
        if os.path.exists(extra_data_path):
            extra_data = read_extra_data(extra_data_path)
        all_data['extra_data'] = extra_data

    # 2. 取出其他settings
    hidden_columns = all_data['hidden_columns']
    column_order = all_data['column_order']
    editable_columns = all_data['editable_columns']
    exclude_columns = all_data['exclude_columns']
    ignore_unchanged_columns = all_data['basic_settings']['ignore_unchanged_columns']
    str_max_length = all_data['basic_settings']['str_max_length']
    round_to = all_data['basic_settings']['round_to']

    new_all_data = generate_columns(logs=logs, hidden_columns=hidden_columns, column_order=column_order,
                                editable_columns=editable_columns,
                                exclude_columns=exclude_columns,
                                filter_condition=all_data['filter_condition'],
                                ignore_unchanged_columns=ignore_unchanged_columns,
                                str_max_length=str_max_length, round_to=round_to,
                                ignore_not_exist=all_data['settings']['Ignore_filter_condition_not_exist_log'])
    all_data.update(new_all_data)

    field_columns = {}
    for key, value in all_data['column_dict'].items():
        if 'field' in value:
            field_columns[key] = 1
    all_data['field_columns'] = field_columns

    return all_data

def replace_with_extra_data(data, extra_data, filter_condition=None, deleted_rows=None, ignore_not_exist=False):
    """

    :param data: {}, key是id，value是一阶json，包含了各个field的值
    :param extra_data: {}, key是id，value是一阶json，包含了各个field的值
    :param filter_condition: {}, 一级json。满足条件才加入(如果对应位置为空，也算满足条件)
    :param deleted_rows:{}, 一级json。在里面的id不能出现在返回的data中
    :param bool ignore_not_exist: 是否忽略不存在的filter_condition的key的log
    :return: 对data进行inplace修改
    """
    # 将数据进行替换
    extra_data = extra_data.copy()
    if len(extra_data)!=0:
        for d, value in data.items():
            if d in extra_data:
                tmp = extra_data.pop(d)
                for k, v in tmp.items():
                    value[k] = v
    # 将新增到extra_data的内容加进去
    if filter_condition is None:
        filter_condition = {}
    if deleted_rows is None:
        deleted_rows = {}
    if len(extra_data)>0: # 还有剩余的，说明是新加入的
        for key, value in extra_data.items(): # key是log的id，value是这个log新加入的内容
            if key in deleted_rows:
                continue
            filter = False  # 是否忽略掉
            for f_k, f_v in filter_condition.items():
                if f_k in value:
                    if str(value[f_k])!=f_v:
                        filter = True
                elif ignore_not_exist:
                    filter = True
            if not filter:
                if 'id' in value:  # 只有有id的才是用户加入的row
                    data[key] = value

def save_all_data(all_data, log_dir, log_config_name):
    # 保存settings和extra文件
    if all_data['settings']['Save_settings']:  # 如果需要保存
        log_config_path = os.path.join(log_dir, log_config_name)
        save_config(all_data, config_path=log_config_path)
        # save editable columns
        if len(all_data['extra_data']) != 0:
            extra_data_path = os.path.join(log_dir, 'log_extra_data.txt')
            save_extra_data(extra_data_path, all_data['extra_data']) # extra_data是一个dict。key为id，value为内容
        print("Settings are saved to {}.".format(log_config_path))



if __name__ == '__main__':
    _dict = json.loads('{"id":0, "a":1,"b":{"a":{"d":4,"e":5},"b":2}}')
    print(_dict)
    data = generate_columns([_dict], column_order={'memo':0, 'b':{'b':0, 'a':{'e':0, 'd':0}}, 'a':0, 'id':0})
    print(data['column_order'])
    print(data['column_dict'])
    import json
    print(json.dumps(_dict))

    # _dict = {'a': 1}
    # print(str(_dict))
    # print(json.loads('{"a":1,"b":2,"c":3,"d":4,"e":5}'))
    # print(type(json.loads('{"a":1,"b":{"a":1,"b":2},"c":3,"d":4,"e":5}')))



from flask import render_template


from flask import request, jsonify
import random
import os
from flask import Blueprint
from .server.data_container import all_data, all_handlers, handler_watcher
from ..fastlog.log_read import is_log_dir_has_step, is_log_record_finish
from .server.chart_utils import ChartStepLogHandler
from .server.utils import replace_nan_inf
from .server.utils import check_uuid
from .server.chart_utils import _refine_logs

import uuid

# 用于进行字符替换，以使得字符串变短
full_name_map = {'name':'n', 'epoch':'e', 'value':'v', 'step': 's'}

chart_page = Blueprint("chart_page", __name__, template_folder='templates')

@chart_page.route('/chart', methods=['POST'])
def chart():
    log_dir = request.values['log_dir']
    finish = request.values['finish']
    save_log_dir = os.path.join(all_data['root_log_dir'], log_dir)
    chart_exclude_columns = all_data['chart_settings']['chart_exclude_columns']
    _uuid = str(uuid.uuid1())
    max_points = all_data['chart_settings']['max_points']
    update_every_second = all_data['chart_settings']['update_every']
    wait_seconds = update_every_second*3 # 如果本来应该收到三次更新，但是却没有收到，则自动关闭
    handler = ChartStepLogHandler(save_log_dir, _uuid, round_to=all_data['basic_settings']['round_to'],
                            max_steps=max_points,
                            wait_seconds=wait_seconds,
                            exclude_columns=chart_exclude_columns,
                            max_no_updates=all_data['chart_settings']['max_no_updates'])
    only_once = is_log_record_finish(save_log_dir) or finish=='true'
    points = handler.update_logs(only_once) # {'loss': [{}, {}], 'metric':[{}, {}]}
    total_steps = points.pop('total_steps', None)
    if not only_once:
        all_handlers[_uuid] = handler
        if not handler_watcher._start:
            handler_watcher.start()

    replace_nan_inf(points)

    return render_template('chart.html', log_dir=log_dir, data=points, chart_uuid=_uuid,
                           max_steps=max_points,
                           server_uuid=all_data['uuid'],
                           update_every=update_every_second*1000,
                           max_no_updates=all_data['chart_settings']['max_no_updates'],
                           total_steps=total_steps,
                           short_name_map = {value:key for key,value in full_name_map.items()})

@chart_page.route('/chart/new_step', methods=['POST'])
def chart_new_step():
    # 获取某个log_dir的更新
    _uuid = request.json['chart_uuid']

    # TODO 使用更好的办法解决一下。
    max_point_per_update = 100  # 一次更新只能选取最新的这么多个数据，否则网络传输和前端显示会有问题
    points = {}
    if _uuid in all_handlers:
        handler = all_handlers[_uuid]
        points = handler.update_logs()
        for key, value in points.items():
            if isinstance(value, list):
                if len(value)<max_point_per_update:
                    continue
                # 只选择max_point_per_update个更新
                total_points = len(value)
                cut_values = [v for v in value if random.random()<max_point_per_update/total_points]
                points[key] = cut_values
    else:
        points['finish'] = True

    replace_nan_inf(points)

    return jsonify(steps=points)

@chart_page.route('/chart/have_trends', methods=['POST'])
def have_trends():
    try:
        res = check_uuid(all_data['uuid'], request.json['uuid'])
        if res != None:
            return jsonify(res)
        log_dir = request.json['log_dir']
        save_log_dir = os.path.join(all_data['root_log_dir'], log_dir)
        if is_log_dir_has_step(save_log_dir):
            return jsonify(status='success', have_trends=True)
        else:
            return jsonify(status='success', have_trends=False, msg='There is no trend data for this log.')
    except Exception:
        print("Exception detected in have_trends()")
        return jsonify(status='fail', have_trends=False, msg='Error from the server.')

@chart_page.route('/chart/range', methods=['POST'])
def ranges():
    try:
        res = check_uuid(all_data['uuid'], request.json['uuid'])
        if res != None:
            return jsonify(res)
        keys = request.json['keys']
        log_dir = request.json['log_dir']
        ranges = request.json['ranges']
        handler = ChartStepLogHandler(save_log_dir=os.path.join(all_data['root_log_dir'], log_dir),
                                      uuid=all_data['uuid'], round_to=all_data['basic_settings']['round_to'],
                                      max_steps=100000,
                                      wait_seconds=10,
                                      exclude_columns=all_data['chart_settings']['chart_exclude_columns'],
                                      max_no_updates=all_data['chart_settings']['max_no_updates'])
        filepaths = []
        for key in keys:
            filepaths.append(os.path.join(all_data['root_log_dir'], log_dir, key+'.log'))
        updates = handler.read_single_update(filepaths, ranges)
        del handler

        refined_updates = {}
        for key in keys:
            logs = updates[key]
            refined_updates[key] = _refine_logs(logs, all_data['chart_settings']['max_points'],
                                                all_data['basic_settings']['round_to'])
        return jsonify(status='success', steps=refined_updates)

    except Exception as e:
        print(e)
        return jsonify(status='fail', msg='Some bug happens, contact developer.')







"""
fastlog 是 fitlog 的一部分，用于帮助用户进行日志管理。fastlog 默认使用工作目录下的 logs 文件做为存储和配置路径。
默认配置文件为default.cfg。用户可以使用 fitlog 命令行工具初始化项目，启动服务器，并在网页上查看日志。
"""
from .log_read import LogReader
from .logger import Logger

logger = Logger()
log_reader = LogReader()



#!/usr/bin/env python
# coding=utf-8
from setuptools import setup, find_packages

with open('README.md', encoding='utf-8') as f:
    readme = f.read()

with open('LICENSE', encoding='utf-8') as f:
    license = f.read()

with open('requirements.txt', encoding='utf-8') as f:
    reqs = f.read()

setup(
    name='fitlog',
    version='0.1.2',
    description='fitlog: Log tool for Deep Learning, developed by Fudan FastNLP Team',
    long_description=readme,
    long_description_content_type='text/markdown',
    license='Apache license',
    python_requires='>=3.6',
    include_package_data=True,
    packages=find_packages(),
    install_requires=reqs.strip().split('\n'),
    entry_points={
        'console_scripts':[
            'fitlog = fitlog.__main__:main_cmd'
        ]
    }
)





import threading
import time

class HandlerWatcher(threading.Thread):
    def __init__(self):
        super().__init__()
        self.all_handlers = all_handlers
        self._stop_flag = False
        self._quit = True
        self._start = False

    def run(self):
        self._quit = False
        self._start = True
        while not self._stop_flag:
            if len(self.all_handlers)>0:
                for _uuid in list(self.all_handlers.keys()):
                    handler = self.all_handlers[_uuid]
                    if handler.reader._quit:
                        handler.reader.stop()
                        handler = self.all_handlers.pop(_uuid)
                        del handler
            time.sleep(0.5)
        # 删除所有的handler
        for _uuid in list(self.all_handlers.keys()):
            handler = self.all_handlers.pop(_uuid)
            if handler.reader._quit:
                handler.reader.stop()
                del handler

        self._quit = True
    def stop(self):
        self._stop_flag = True
        count = 0
        while not self._quit:
            time.sleep(0.6)
            if count>3:
                raise RuntimeError("Some bug happens.")
            count += 1
# singleton
"""
all_data包含以下的key:
    settings: {} 一级dict，包含了所有的frontend_settings中的内容，value全部是bool值
    basic_settings: {} 一级dict包含了config中basic_settings中的setting.
    hidden_rows: {} 一级dict key为隐藏的row的id
    deleted_rows: {} 一级dict key为删除的row的id
    filter_condition: {} 一级dict，expanded的key以及它等于的value
    hidden_columns: {} 一级dict key为隐藏的column
    exclude_columns: {} 一级dict，需要排除的column
    editable_columns: {} 一级dict，支持编辑的column名
    column_order: {} nested dict. 表示column的顺序的
    chart_settings: {} 保存chart相关的设置，包含以下的内容
        chart_exclude_columns:{} 一级dict，需要排除的column名称
        max_points:int 前端每条线最多显示多少个点
        update_every: int, 隔多少秒update一次
        max_no_updates: int 多少次没有得到更新就认为已经停止了
    config: 读取的ConfigParser对象
    extra_data: {}, 第一层的key为前端增加的记录获取用户修改某条记录留下的修改记录; 第一层的value对应的是一个一级dict。
    root_log_dir: str, log文件夹的路径
    log_config_name: str, 
    log_reader: LogReader()对象
    port: int, port
    uuid: str, 这个server的uuid
    token: str,None 这个server的token，放访问路径上的
    data: {id1:{'id':id1, 'field1':xxx,}, id2:{}}, 所有的数据都在这个里面，这是一个一级dict
"""
all_data = {}
all_handlers = {}
handler_watcher = HandlerWatcher()


from flask import render_template

from flask import request, jsonify
from flask import Blueprint
import os
import shutil

from .server.table_utils import prepare_data, prepare_incremental_data

from .server.data_container import all_data
from ..fastgit import committer
from .server.utils import replace_nan_inf
from .server.utils import check_uuid
from .server.table_utils import save_all_data
from .server.table_utils import replace_with_extra_data

from werkzeug.utils import secure_filename
from .server.server_config import _get_config_names

table_page = Blueprint("table_page", __name__, template_folder='templates')

first_time_access = True

@table_page.route('/table/table')
def get_table():
    global first_time_access
    if not first_time_access:
        log_dir = all_data['root_log_dir']
        log_config_name = all_data['log_config_name']
        log_reader = all_data['log_reader']
        log_reader.set_log_dir(log_dir)
        all_data.update(prepare_data(log_reader, log_dir, log_config_name, all_data))

    first_time_access = False
    data = all_data['data'].copy() # all_data['data'] 只包含从硬盘的读取的log的信息
    replace_with_extra_data(data, all_data['extra_data'], all_data['filter_condition'], all_data['deleted_rows'],
                             all_data['settings']['Ignore_filter_condition_not_exist_log'])
    replace_nan_inf(data)

    return jsonify(column_order=all_data['column_order'], column_dict=all_data['column_dict'],
                   hidden_columns=all_data['hidden_columns'],
                   data=data,
                   settings={key.replace('_', ' '):value for key, value in all_data['settings'].items()},
                   uuid=all_data['uuid'],
                   hidden_rows=list(all_data['hidden_rows'].keys()),
                   unchanged_columns=all_data['unchanged_columns'],
                   log_config_name=all_data['log_config_name'])

@table_page.route('/table/refresh', methods=['POST'])
def refresh_table():
    res = check_uuid(all_data['uuid'], request.json['uuid'])
    if res != None:
        return jsonify(res)
    log_reader = all_data['log_reader']
    new_logs = log_reader.read_logs(all_data['deleted_rows'])
    # 删除不满足条件的
    try:
        if len(new_logs)==0:
            return jsonify(status='success', msg='Update successfully, no update found.', new_logs=[], updated_logs=[])
        else:
            new_logs, updated_logs = prepare_incremental_data(all_data['data'], new_logs, all_data['field_columns'],
                                                              all_data['filter_condition'],
                                                              all_data['settings'][
                                                                  'Ignore_filter_condition_not_exist_log'])
            if len(new_logs)==0 and len(updated_logs)==0:
                return jsonify(status='success', msg='Update successfully, no update found.', new_logs=[],
                               updated_logs=[])
            # 需要过滤filter_condition
            replace_nan_inf(new_logs)
            replace_nan_inf(updated_logs)
            return jsonify(status='success', msg='Update successfully, {} log have updates, {} newly added.'\
                           .format(len(updated_logs), len(new_logs)),
                           new_logs=new_logs, updated_logs=updated_logs)
    except:
        return jsonify(status='fail', msg="Unknown error from server.")

@table_page.route('/table/delete_records', methods=['POST'])
def delete_records():
    res = check_uuid(all_data['uuid'], request.json['uuid'])
    if res != None:
        return jsonify(res)
    ids = request.json['ids']
    for id in ids:
        if id in all_data['data']:
            all_data['deleted_rows'][id] = 1
        elif id in all_data['extra_data']:
            # all_data['extra_data'].pop(id)
            all_data['deleted_rows'][id] = 1

    return jsonify(status='success', msg='')

@table_page.route('/table/erase_records', methods=['POST'])
def erase_records():
    res = check_uuid(all_data['uuid'], request.json['uuid'])
    if res != None:
        return jsonify(res)
    ids = request.json['ids']
    fail_ids = []
    for index in range(len(ids)-1, -1, -1):
        id = ids[index]
        if id in all_data['data']:
            all_data['data'].pop(id, None)
            all_data['deleted_rows'].pop(id, None)
            # 删除
            try:
                record_path = os.path.join(all_data['root_log_dir'], id)
                if os.path.isdir(record_path):
                    shutil.rmtree(record_path)
            except Exception as e:
                fail_ids.append(id)

        elif id in all_data['extra_data']:
            all_data['extra_data'].pop(id)

    if len(fail_ids)!=0:
        return jsonify(status='fail', msg=fail_ids)
    else:
        return jsonify(status='success', msg='')

@table_page.route('/table/edit', methods=['POST'])
def table_edit():
    try:
        # 包含field, id, new_field_value三个值. 修改某个值
        res = check_uuid(all_data['uuid'], request.json['uuid'])
        if res != None:
            return jsonify(res)
        id = request.json['id']
        field = request.json['field']
        new_field_value = request.json['new_field_value']
        if id in all_data['extra_data']:
            all_data['extra_data'][id][field] = new_field_value
        else:
            all_data['extra_data'][id] = {field:new_field_value}
        if id in all_data['data'] and field in all_data['data'][id]:
            all_data['data'][id][field] = new_field_value

        return jsonify(status='success', msg='')
    except Exception as e:
        print(e)
        return jsonify(status='fail', msg='Unknown error fail to save edit results.')

@table_page.route('/table/reset', methods=['POST'])
def table_reset():
    try:
        res = check_uuid(all_data['uuid'], request.json['uuid'])
        if res != None:
            return jsonify(res)
        fit_id = request.json['fit_id']
        _suffix = request.json['suffix']
        response = committer.fitlog_revert(fit_id, all_data['root_log_dir'], _suffix)
        if response['status'] == 0:
            return jsonify(status='success', msg=response['msg'])
        else:
            return jsonify(status='fail', msg=response['msg'])
    except Exception as e:
        print(e)
        return jsonify(status='fail', msg='Unknown error from server.')

@table_page.route('/table/settings', methods=['POST'])
def settings():
    res = check_uuid(all_data['uuid'], request.json['uuid'])
    if res!=None:
        return jsonify(res)
    settings = request.json['settings']
    all_data['settings'].update({key.replace(' ', '_'):value for key, value in settings.items()})

    return jsonify(status='success', msg='')

@table_page.route('/table/save_config_name', methods=['POST'])
def save_config_name():
    res = check_uuid(all_data['uuid'], request.json['uuid'])
    if res!=None:
        return jsonify(res)
    log_config_name = request.json['save_config_name']
    log_config_name = secure_filename(log_config_name)
    if len(log_config_name)!=0:
        if all_data['log_config_name']!=log_config_name:
            all_data['log_config_name'] = log_config_name
        return jsonify(status='success', msg=log_config_name)
    else:
        return jsonify(status='fail', msg='Invalid file name')

@table_page.route('/table/hidden_rows', methods=['POST'])
def hidden_ids():
    res = check_uuid(all_data['uuid'], request.json['uuid'])
    if res!=None:
        return jsonify(res)
    ids = request.json['ids']
    all_data['hidden_rows'].clear()
    for id in ids:
        all_data['hidden_rows'][id] = 1

    return jsonify(status='success', msg='')

@table_page.route('/table/hidden_columns', methods=['POST'])
def hidden_columns():
    res = check_uuid(all_data['uuid'], request.json['uuid'])
    if res!=None:
        return jsonify(res)
    hidden_columns = request.json['hidden_columns']
    all_data['hidden_columns'] = hidden_columns
    return jsonify(status='success', msg='')

@table_page.route('/table/column_order', methods=['POST'])
def column_order():
    res = check_uuid(all_data['uuid'], request.json['uuid'])
    if res != None:
        return jsonify(res)
    column_order = request.json['column_order']
    all_data['column_order'] = column_order
    return jsonify(status='success', msg='')

@table_page.route('/table/row', methods=['POST'])
def add_row():
    res = check_uuid(all_data['uuid'], request.json['uuid'])
    if res != None:
        return jsonify(res)
    new_row = request.json['row']
    if new_row['id'] not in all_data['data']:
        all_data['data'][new_row['id']] = new_row
        all_data['extra_data'][new_row['id']] = new_row
        return jsonify(status='success', msg='')
    else:
        return jsonify(status='fail', msg='Duplicated id.')

@table_page.route('/table/save_settings', methods=['POST'])
def save():
    res = check_uuid(all_data['uuid'], request.json['uuid'])
    if res != None:
        return jsonify(res)
    if 'condition' in request.json:
        condition = request.json['condition'] # {'key': 'value'}
        all_data['filter_condition'].update(condition)
    log_dir = all_data['root_log_dir']
    log_config_name = all_data['log_config_name']
    save_all_data(all_data, log_dir, log_config_name)
    return jsonify(status='success', msg='')

@table_page.route('/table')
def table():
    if all_data['token'] != None:
        return jsonify(msg='This url needs a token to access. If you did not specify a token when start the server, '
                           'you may access the wrong ip or port. If you specify a token when start the server, '
                           'use http://{your server ip}:{port}/table/{token} to access.')
    return render_template('table.html')

@table_page.route('/table/<token>')
def table_with_token(token):
    if token!=all_data['token'] or all_data['token'] is None:
        return jsonify(msg='Wrong token or you are not using a token')
    return render_template('table.html')

@table_page.route('/table/configs', methods=['POST'])
def table_configs():
    # 显示有哪些config文件的接口
    res = check_uuid(all_data['uuid'], request.json['uuid'])
    if res != None:
        return jsonify(res)
    # 读取当前可以使用的config有哪些
    config_names = _get_config_names(all_data['root_log_dir'])
    if len(config_names)==0:
        return jsonify(status='fail', msg='There is no valid config file in {}.'.format(all_data['root_log_dir']))
    if config_names.index(all_data['log_config_name'])==-1:
        return jsonify(status='fail', msg='There current config file:{} is not exist.'.format(all_data['log_config_name']))
    configs = {}
    for config in config_names:
        configs[config] = config==all_data['log_config_name']
    return jsonify(status='success', msg='', configs=configs)

@table_page.route('/table/change_config', methods=['POST'])
def table_change_config():
    # 改变config的接口
    res = check_uuid(all_data['uuid'], request.json['uuid'])
    if res != None:
        return jsonify(res)
    if 'config_name' in request.json:
        config_names = _get_config_names(all_data['root_log_dir'])
        if config_names.index(request.json['config_name'])!=-1:
            log_dir = all_data['root_log_dir']
            log_config_name = all_data['log_config_name']
            save_all_data(all_data, log_dir, log_config_name)
            all_data['log_config_name'] = request.json['config_name']
            return jsonify(status='success', msg='')
        else:
            return jsonify(status='fail', msg='There is no config named:{}.'.format(request.json['config_name']))
    else:
        return jsonify(status='fail', msg='There is no config_name in your request.')



import os
import uuid

from flask import Flask, url_for, redirect
from .chart_app import chart_page
from .table_app import table_page
from .server.table_utils import save_all_data
from .server.app_utils import get_usage_port
from .server.data_container import all_data
from .server.data_container import handler_watcher
from ..fastlog import log_reader
from .server.table_utils import prepare_data
from flask import request
import time
from .server.app_utils import ServerWatcher
from flask import send_from_directory
from urllib import request as urequest
from threading import Timer
from flask import jsonify
from .server.utils import check_uuid

from collections import deque

app = Flask(__name__)

app.register_blueprint(chart_page)
app.register_blueprint(table_page)

LEAST_REQUEST_TIMESTAMP = deque(maxlen=1)
LEAST_REQUEST_TIMESTAMP.append(time.time())

server_watcher = ServerWatcher(LEAST_REQUEST_TIMESTAMP)

@app.route('/')
def index():
    return redirect(url_for('table_page.table'))

@app.before_request
def update_last_request_ms():
    global LEAST_REQUEST_TIMESTAMP
    LEAST_REQUEST_TIMESTAMP.append(time.time())

@app.route('/kill', methods=['POST'])
def seriouslykill():
    time.sleep(1)
    func = request.environ.get('werkzeug.server.shutdown')
    if func is None:
        raise RuntimeError('Not running with the Werkzeug Server')
    func()
    return "stopping"

@app.route('/arange_kill', methods=['POST'])
def arange_kill():
    res = check_uuid(all_data['uuid'], request.json['uuid'])
    if res!=None:
        return jsonify(res)
    def shutdown():
        req = urequest.Request('http://127.0.0.1:{}/kill'.format(all_data['port']), headers={}, data=''.encode('utf-8'))
        page = urequest.urlopen(req).read().decode('utf-8')
    print("Shutting down from the frontend...")
    Timer(1.0, shutdown).start()
    return jsonify(status='success', msg='')

@app.route('/table.ico')
def get_table_ico():
    return send_from_directory(os.path.join('.', 'static', 'img'), 'table.ico')

@app.route('/chart.ico')
def get_chart_ico():
    return send_from_directory(os.path.join('.', 'static', 'img'), 'chart.ico')

def start_app(log_dir, log_config_name, start_port, standby_hours, token=None):
    os.chdir(os.path.dirname(os.path.abspath(__file__))) # 可能需要把运行路径移动到这里
    all_data['root_log_dir'] = log_dir # will be used by chart_app
    server_wait_seconds = int(standby_hours*3600)
    all_data['log_config_name'] = log_config_name
    log_reader.set_log_dir(log_dir)
    all_data['log_reader'] = log_reader
    if token is None:
        all_data['token'] = None
    else:
        all_data['token'] = str(token)

    # 准备数据
    all_data.update(prepare_data(log_reader, all_data['root_log_dir'], all_data['log_config_name']))
    print("Finish preparing data. Found {} records in {}.".format(len(all_data['data']), log_dir))
    all_data['uuid'] = str(uuid.uuid1())

    port = get_usage_port(start_port=start_port)
    server_watcher.set_server_wait_seconds(server_wait_seconds)
    server_watcher.start()
    if all_data['token']!=None:
        print("You specify token:{}, remember to add this token when access your table.".format(all_data['token']))
    all_data['port'] = port
    app.run(host='0.0.0.0', port=port, debug=False, threaded=True)

    # TODO 输出访问的ip地址
    print("Shutting down server...")
    save_all_data(all_data,  all_data['root_log_dir'], all_data['log_config_name'])
    handler_watcher.stop()
    server_watcher.stop()

if __name__ == '__main__':
    from .server.app_utils import cmd_parser
    parser = cmd_parser()
    args =parser.parse_args()

    start_app(args.log_dir, args.log_config_name, args.port, 1, '123')







from .log_config_parser import ConfigParser
import os
import json
import glob

def read_server_config(config_path):
    """
    给定config的path，读取里面的config。如果config不存在，则按照默认的值创建
    :param config_path: str
    :return: dict, config的内容
    """
    config = ConfigParser(allow_no_value=True)
    if not os.path.exists(config_path):
        config_dir = os.path.dirname(config_path)
        if not os.path.isdir(config_dir):
            os.makedirs(config_dir)
        _read_default_config(config)
        with open(config_path, 'w') as f:
            config.write(f)
    else:
        config.read(config_path)
        # check configparser, 没有的话用默认值覆盖填写
        check_config(config)

    all_data = {}
    # 读取settings
    settings = {}
    get_dict_from_config(config, 'frontend_settings', settings, 'bool')
    settings = {key.capitalize():value for key, value in settings.items()}
    all_data['settings'] = settings

    # 读取display_settings
    basic_settings = {}
    get_dict_from_config(config, 'basic_settings', basic_settings)
    for key in ['str_max_length', 'round_to']:
        basic_settings[key] = int(basic_settings[key])
    for key in ['ignore_unchanged_columns']:
        basic_settings[key] = basic_settings[key]=='True'
    all_data['basic_settings'] = basic_settings


    # 读取data_settings
    hidden_logs = read_list_from_config(config, 'data_settings', 'hidden_logs', ',')
    all_data['hidden_rows'] = {log:1 for log in hidden_logs}
    deleted_rows = read_list_from_config(config, 'data_settings', 'deleted_logs', ',')
    all_data['deleted_rows'] = {log: 1 for log in deleted_rows}
    if len(config.get('data_settings', 'filter_condition'))!=0:
        all_data['filter_condition'] = json.loads(config.get('data_settings', 'filter_condition'))
    else:
        all_data['filter_condition'] = {}

    # 读取column_settings
    hidden_columns = read_list_from_config(config, 'column_settings', 'hidden_columns', ',')
    all_data['hidden_columns'] = {column:1 for column in hidden_columns}

    # 读取exclude columns
    exclude_columns = read_list_from_config(config, 'column_settings', 'exclude_columns', ',')
    all_data['exclude_columns'] = {column:1 for column in exclude_columns}

    # 读取editable_columns
    editable_columns = read_list_from_config(config, 'column_settings', 'editable_columns', ',')
    all_data['editable_columns'] = {column:1 for column in editable_columns}

    # 读取column_order
    column_order = config.get('column_settings', 'column_order')
    if column_order!='':
        column_order = json.loads(column_order)
    else:
        column_order = {}
    all_data['column_order'] = column_order

    # 保存config对象, 因为需要保留下注释
    all_data['config'] = config

    _dict = {}
    chart_exclude_columns = read_list_from_config(config, 'chart_settings', 'chart_exclude_columns', ',')
    _dict['chart_exclude_columns'] = {column:1 for column in chart_exclude_columns}
    _dict['max_points'] = config.getint('chart_settings', 'max_points')
    _dict['update_every'] = config.getint('chart_settings', 'update_every')
    _dict['max_no_updates'] = config.getint('chart_settings', 'max_no_updates')
    all_data['chart_settings'] = _dict

    return all_data

def save_config(all_data, config_path):
    config = all_data['config']
    if not os.path.exists(config_path):
        config_dir = os.path.dirname(config_path)
        if not os.path.isdir(config_dir):
            os.makedirs(config_dir)
        os.mknod(config_path)  # 不存在就创建空文件

    # frontend_settings
    settings = all_data['settings']
    settings = {key.replace(' ', '_'):value for key, value in settings.items()}
    save_dict_to_config(config, 'frontend_settings', settings)

    # basic_settings
    basic_settings = all_data['basic_settings']
    save_dict_to_config(config, 'basic_settings', basic_settings)

    # data settings
    hidden_logs = all_data['hidden_rows'].keys()
    config.set('data_settings', 'hidden_logs', ','.join(hidden_logs))
    deleted_logs = all_data['deleted_rows'].keys()
    config.set('data_settings', 'deleted_logs', ','.join(deleted_logs))
    if len(all_data['filter_condition'])!=0:
        filter_condition = json.dumps(all_data['filter_condition'])
    else:
        filter_condition = ''
    config.set('data_settings', 'filter_condition', filter_condition)

    # column_settings
    for option in ['hidden_columns', 'editable_columns']:
        save_list_to_config(config, 'column_settings', option, all_data[option])
    column_order = refine_column_order(all_data['column_order'])
    column_order = json.dumps(column_order)
    config.set('column_settings', 'column_order', column_order)

    with open(config_path, 'w', encoding='utf-8') as f:
        config.write(f)

def refine_column_order(column_order):
    # 重新生成column_order
    new_column_order = {}
    if 'OrderKeys' in column_order: # 还有下一级
        keys = column_order['OrderKeys']
        for key in keys:
            value = column_order[key]
            child = refine_column_order(value)
            if len(child)==0:
                new_column_order[key] = 'EndOfOrder'
            else:
                new_column_order[key] = child
    return new_column_order


def save_list_to_config(config, section, option, container, sep=','):
    if not config.has_section(section):
        config.add_section(section)
    if len(container)==0:
        str_ = ''
    else:
        str_ = sep.join(container.keys())
    config.set(section, option, str_)


def save_dict_to_config(config, section, container):
    if not config.has_section(section):
        config.add_section(section)
    for key, value in container.items():
        config.set(section, key, str(value))


def read_list_from_config(config, section, option, sep):
    str_ = config.get(section, option)
    items = []
    if str_!='':
        items = str_.split(sep)
        items = [item.strip() for item in items]
    return items

def _read_default_config(config):
    """

    :param config: ConfigParser对象
    :return:
    """
    default_cfg_fp = os.path.join(os.path.realpath(__file__)[:-len("server_config.py")], '..',
                                  '..', 'fastgit', 'normal', 'logs', 'default.cfg')
    config.read(default_cfg_fp)
    return config

def check_config(config):
    # 检查config是否拥有所有的值，如果没有的话，使用默认值填写
    default_config = ConfigParser()
    default_config = _read_default_config(default_config)
    for section in default_config.sections():
        if not config.has_section(section):
            # raise KeyError("Section:`{}` is not in config.".format(section))
            config.add_section(section)
            for option, value in default_config.items(section):
                config.set(section, option, value)
        for opt in default_config.options(section):
            if not config.has_option(section, opt):
                # raise KeyError("Option:`{}` is not in Section:`{}` is not in config.".format(opt, section))
                config.set(section, opt, default_config.get(section, opt))

def get_dict_from_config(config, section, container, dtype=None):
    if dtype == 'int':
        func = config.getint
    elif dtype == 'bool' or dtype=='boolean':
        func = config.getboolean
    else:
        func = config.get
    for option in config.options(section):
        container[option] = func(section, option)

def read_extra_data(path):
    with open(path, 'r', encoding='utf-8') as f:
        data = json.load(f)
    return data

def save_extra_data(path, data):
    with open(path, 'w', encoding='utf-8') as f:
        json.dump(data, f)


def _get_config_names(root_log_dir):
    """
    给定log的路径, 返回下面所有结尾为.cfg的文件

    :param str, root_log_dir:
    :return: list(dir)
    """
    configs = glob.glob(os.path.join(root_log_dir, '*.cfg'))
    configs = [os.path.basename(config) for config in configs]
    return configs


